from pathlib import Path
from .base import ModuleTestBase, tempapkfile


class TestAPKPure(ModuleTestBase):
    modules_overrides = ["apkpure", "google_playstore", "speculate"]
    apk_file = tempapkfile()

    async def setup_after_prep(self, module_test):
        await module_test.mock_dns({"blacklanternsecurity.com": {"A": ["127.0.0.99"]}})
        module_test.httpx_mock.add_response(
            url="https://play.google.com/store/search?q=blacklanternsecurity&c=apps",
            text="""<!DOCTYPE html>
            <html>
            <head>
            <title>"blacklanternsecurity" - Android Apps on Google Play</title>
            </head>
            <body>
            <a href="/store/apps/details?id=com.bbot.test&pcampaignid=dontmatchme&pli=1"/>
            </body>
            </html>""",
        )
        module_test.httpx_mock.add_response(
            url="https://play.google.com/store/apps/details?id=com.bbot.test",
            text="""<!DOCTYPE html>
            <html>
            <head>
            <title>BBOT</title>
            </head>
            <body>
            <meta name="appstore:developer_url" content="https://www.blacklanternsecurity.com">
            </div>
            </div>
            </body>
            </html>""",
        )
        module_test.httpx_mock.add_response(
            url="https://d.apkpure.com/b/XAPK/com.bbot.test?version=latest",
            content=self.apk_file,
        )

    def check(self, module_test, events):
        assert len(events) == 6
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "blacklanternsecurity.com" and e.scope_distance == 0
            ]
        ), "Failed to emit target DNS_NAME"
        assert 1 == len(
            [e for e in events if e.type == "ORG_STUB" and e.data == "blacklanternsecurity" and e.scope_distance == 0]
        ), "Failed to find ORG_STUB"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "MOBILE_APP"
                and "android" in e.tags
                and e.data["id"] == "com.bbot.test"
                and e.data["url"] == "https://play.google.com/store/apps/details?id=com.bbot.test"
            ]
        ), "Failed to find bbot android app"
        filesystem_event = [
            e for e in events if e.type == "FILESYSTEM" and "com.bbot.test.xapk" in e.data["path"] and "apk" in e.tags
        ]
        assert 1 == len(filesystem_event), "Failed to download apk"
        file = Path(filesystem_event[0].data["path"])
        assert file.is_file(), "Destination xapk doesn't exist"



from bbot.core.helpers.dns.helpers import common_srvs
from bbot.modules.templates.subdomain_enum import subdomain_enum


class dnscommonsrv(subdomain_enum):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "active", "safe"]
    meta = {"description": "Check for common SRV records", "created_date": "2022-05-15", "author": "@TheTechromancer"}
    dedup_strategy = "lowest_parent"

    options = {"max_depth": 2}
    options_desc = {"max_depth": "The maximum subdomain depth to brute-force SRV records"}

    async def setup(self):
        self.max_subdomain_depth = self.config.get("max_depth", 2)
        self.num_srvs = len(common_srvs)
        return True

    async def filter_event(self, event):
        subdomain_depth = self.helpers.subdomain_depth(event.host)
        if subdomain_depth > self.max_subdomain_depth:
            return False, f"its subdomain depth ({subdomain_depth}) exceeds max_depth={self.max_subdomain_depth}"
        return True

    async def handle_event(self, event):
        query = self.make_query(event)
        self.verbose(f'Brute-forcing {self.num_srvs:,} SRV records for "{query}"')
        for hostname in await self.helpers.dns.brute(self, query, common_srvs, type="SRV"):
            await self.emit_event(
                hostname,
                "DNS_NAME",
                parent=event,
                context=f'{{module}} tried {self.num_srvs:,} common SRV records against "{query}" and found {{event.type}}: {{event.data}}',
            )



import asyncio
import datetime
import ipaddress

from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_helpers_misc(helpers, scan, bbot_scanner, bbot_httpserver):
    ### URL ###
    bad_urls = (
        "http://e.co/index.html",
        "http://e.co/u/1111/info",
        "http://e.co/u/2222/info",
        "http://e.co/u/3333/info",
        "http://e.co/u/4444/info",
        "http://e.co/u/5555/info",
    )
    new_urls = tuple(helpers.validators.collapse_urls(bad_urls, threshold=4))
    assert len(new_urls) == 2
    new_urls = tuple(sorted([u.geturl() for u in helpers.validators.collapse_urls(bad_urls, threshold=5)]))
    assert new_urls == bad_urls

    new_url = helpers.add_get_params("http://evilcorp.com/a?p=1&q=2", {"r": 3, "s": "asdf"}).geturl()
    query = dict(s.split("=") for s in new_url.split("?")[-1].split("&"))
    query = tuple(sorted(query.items(), key=lambda x: x[0]))
    assert query == (
        ("p", "1"),
        ("q", "2"),
        ("r", "3"),
        ("s", "asdf"),
    )
    assert tuple(sorted(helpers.get_get_params("http://evilcorp.com/a?p=1&q=2#frag").items())) == (
        ("p", ["1"]),
        ("q", ["2"]),
    )

    assert helpers.validators.clean_url("http://evilcorp.com:80").geturl() == "http://evilcorp.com/"
    assert helpers.validators.clean_url("http://evilcorp.com/asdf?a=asdf#frag").geturl() == "http://evilcorp.com/asdf"
    assert helpers.validators.clean_url("http://evilcorp.com//asdf").geturl() == "http://evilcorp.com/asdf"
    assert helpers.validators.clean_url("http://evilcorp.com.").geturl() == "http://evilcorp.com/"
    with pytest.raises(ValueError):
        helpers.validators.clean_url("http://evilcorp,com")

    assert helpers.url_depth("http://evilcorp.com/asdf/user/") == 2
    assert helpers.url_depth("http://evilcorp.com/asdf/user") == 2
    assert helpers.url_depth("http://evilcorp.com/asdf/") == 1
    assert helpers.url_depth("http://evilcorp.com/asdf") == 1
    assert helpers.url_depth("http://evilcorp.com/") == 0
    assert helpers.url_depth("http://evilcorp.com") == 0

    assert helpers.parent_url("http://evilcorp.com/subdir1/subdir2?foo=bar") == "http://evilcorp.com/subdir1"

    ### MISC ###
    assert helpers.is_domain("evilcorp.co.uk")
    assert not helpers.is_domain("www.evilcorp.co.uk")
    assert helpers.is_domain("evilcorp.notreal")
    assert not helpers.is_domain("asdf.evilcorp.notreal")
    assert not helpers.is_domain("notreal")
    assert helpers.is_subdomain("www.evilcorp.co.uk")
    assert not helpers.is_subdomain("evilcorp.co.uk")
    assert helpers.is_subdomain("www.evilcorp.notreal")
    assert not helpers.is_subdomain("evilcorp.notreal")
    assert not helpers.is_subdomain("notreal")
    assert helpers.is_url("http://evilcorp.co.uk/asdf?a=b&c=d#asdf")
    assert helpers.is_url("https://evilcorp.co.uk/asdf?a=b&c=d#asdf")
    assert helpers.is_uri("ftp://evilcorp.co.uk") == True
    assert helpers.is_uri("http://evilcorp.co.uk") == True
    assert helpers.is_uri("evilcorp.co.uk", return_scheme=True) == ""
    assert helpers.is_uri("ftp://evilcorp.co.uk", return_scheme=True) == "ftp"
    assert helpers.is_uri("FTP://evilcorp.co.uk", return_scheme=True) == "ftp"
    assert not helpers.is_url("https:/evilcorp.co.uk/asdf?a=b&c=d#asdf")
    assert not helpers.is_url("/evilcorp.co.uk/asdf?a=b&c=d#asdf")
    assert not helpers.is_url("ftp://evilcorp.co.uk")
    assert helpers.parent_domain("www.evilcorp.co.uk") == "evilcorp.co.uk"
    assert helpers.parent_domain("evilcorp.co.uk") == "evilcorp.co.uk"
    assert helpers.parent_domain("localhost") == "localhost"
    assert helpers.parent_domain("www.evilcorp.notreal") == "evilcorp.notreal"
    assert helpers.parent_domain("evilcorp.notreal") == "evilcorp.notreal"
    assert helpers.parent_domain("notreal") == "notreal"
    assert list(helpers.domain_parents("test.www.evilcorp.co.uk")) == ["www.evilcorp.co.uk", "evilcorp.co.uk"]
    assert list(helpers.domain_parents("www.evilcorp.co.uk", include_self=True)) == [
        "www.evilcorp.co.uk",
        "evilcorp.co.uk",
    ]
    assert list(helpers.domain_parents("evilcorp.co.uk", include_self=True)) == ["evilcorp.co.uk"]
    assert list(helpers.ip_network_parents("0.0.0.0/2")) == [
        ipaddress.ip_network("0.0.0.0/1"),
        ipaddress.ip_network("0.0.0.0/0"),
    ]
    assert list(helpers.ip_network_parents("0.0.0.0/1", include_self=True)) == [
        ipaddress.ip_network("0.0.0.0/1"),
        ipaddress.ip_network("0.0.0.0/0"),
    ]
    assert helpers.is_ip("127.0.0.1")
    assert not helpers.is_ip("127.0.0.0.1")

    assert not helpers.is_ip_type("127.0.0.1")
    assert helpers.is_ip_type(ipaddress.ip_address("127.0.0.1"))
    assert not helpers.is_ip_type(ipaddress.ip_address("127.0.0.1"), network=True)
    assert helpers.is_ip_type(ipaddress.ip_address("127.0.0.1"), network=False)
    assert helpers.is_ip_type(ipaddress.ip_network("127.0.0.0/8"))
    assert helpers.is_ip_type(ipaddress.ip_network("127.0.0.0/8"), network=True)
    assert not helpers.is_ip_type(ipaddress.ip_network("127.0.0.0/8"), network=False)

    assert helpers.is_dns_name("evilcorp.com")
    assert helpers.is_dns_name("evilcorp")
    assert not helpers.is_dns_name("evilcorp", include_local=False)
    assert helpers.is_dns_name("ドメイン.テスト")
    assert not helpers.is_dns_name("127.0.0.1")
    assert not helpers.is_dns_name("dead::beef")
    assert not helpers.is_dns_name("bob@evilcorp.com")

    assert helpers.domain_stem("evilcorp.co.uk") == "evilcorp"
    assert helpers.domain_stem("www.evilcorp.co.uk") == "www.evilcorp"

    assert tuple(await helpers.re.extract_emails("asdf@asdf.com\nT@t.Com&a=a@a.com__ b@b.com")) == (
        "asdf@asdf.com",
        "t@t.com",
        "a@a.com",
        "b@b.com",
    )

    assert helpers.extract_host("evilcorp.com:80") == ("evilcorp.com", "", ":80")
    assert helpers.extract_host("http://evilcorp.com:80/asdf.php?a=b") == (
        "evilcorp.com",
        "http://",
        ":80/asdf.php?a=b",
    )
    assert helpers.extract_host("http://evilcorp.com:80/asdf.php?a=b@a.com") == (
        "evilcorp.com",
        "http://",
        ":80/asdf.php?a=b@a.com",
    )
    assert helpers.extract_host("bob@evilcorp.com") == ("evilcorp.com", "bob@", "")
    assert helpers.extract_host("[dead::beef]:22") == ("dead::beef", "[", "]:22")
    assert helpers.extract_host("scp://[dead::beef]:22") == ("dead::beef", "scp://[", "]:22")
    assert helpers.extract_host("https://[dead::beef]:22?a=b") == ("dead::beef", "https://[", "]:22?a=b")
    assert helpers.extract_host("https://[dead::beef]/?a=b") == ("dead::beef", "https://[", "]/?a=b")
    assert helpers.extract_host("https://[dead::beef]?a=b") == ("dead::beef", "https://[", "]?a=b")
    assert helpers.extract_host("ftp://username:password@my-ftp.com/my-file.csv") == (
        "my-ftp.com",
        "ftp://username:password@",
        "/my-file.csv",
    )
    assert helpers.extract_host("ftp://username:p@ssword@my-ftp.com/my-file.csv") == (
        "my-ftp.com",
        "ftp://username:p@ssword@",
        "/my-file.csv",
    )
    assert helpers.extract_host("ftp://username:password:/@my-ftp.com/my-file.csv") == (
        "my-ftp.com",
        "ftp://username:password:/@",
        "/my-file.csv",
    )
    assert helpers.extract_host("ftp://username:password:/@dead::beef/my-file.csv") == (
        None,
        "ftp://username:password:/@dead::beef/my-file.csv",
        "",
    )
    assert helpers.extract_host("ftp://username:password:/@[dead::beef]/my-file.csv") == (
        "dead::beef",
        "ftp://username:password:/@[",
        "]/my-file.csv",
    )
    assert helpers.extract_host("ftp://username:password:/@[dead::beef]:22/my-file.csv") == (
        "dead::beef",
        "ftp://username:password:/@[",
        "]:22/my-file.csv",
    )

    assert helpers.best_http_status(200, 404) == 200
    assert helpers.best_http_status(500, 400) == 400
    assert helpers.best_http_status(301, 302) == 301
    assert helpers.best_http_status(0, 302) == 302
    assert helpers.best_http_status(500, 0) == 500

    assert helpers.split_domain("www.evilcorp.co.uk") == ("www", "evilcorp.co.uk")
    assert helpers.split_domain("asdf.www.test.notreal") == ("asdf.www", "test.notreal")
    assert helpers.split_domain("www.test.notreal") == ("www", "test.notreal")
    assert helpers.split_domain("test.notreal") == ("", "test.notreal")
    assert helpers.split_domain("notreal") == ("", "notreal")
    assert helpers.split_domain("192.168.0.1") == ("", "192.168.0.1")
    assert helpers.split_domain("dead::beef") == ("", "dead::beef")

    assert helpers.subdomain_depth("a.s.d.f.evilcorp.co.uk") == 4
    assert helpers.subdomain_depth("a.s.d.f.evilcorp.com") == 4
    assert helpers.subdomain_depth("evilcorp.com") == 0
    assert helpers.subdomain_depth("a.evilcorp.com") == 1
    assert helpers.subdomain_depth("a.s.d.f.evilcorp.notreal") == 4

    assert helpers.split_host_port("http://evilcorp.co.uk") == ("evilcorp.co.uk", 80)
    assert helpers.split_host_port("https://evilcorp.co.uk") == ("evilcorp.co.uk", 443)
    assert helpers.split_host_port("ws://evilcorp.co.uk") == ("evilcorp.co.uk", 80)
    assert helpers.split_host_port("wss://evilcorp.co.uk") == ("evilcorp.co.uk", 443)
    assert helpers.split_host_port("WSS://evilcorp.co.uk") == ("evilcorp.co.uk", 443)
    assert helpers.split_host_port("http://evilcorp.co.uk:666") == ("evilcorp.co.uk", 666)
    assert helpers.split_host_port("evilcorp.co.uk:666") == ("evilcorp.co.uk", 666)
    assert helpers.split_host_port("evilcorp.co.uk") == ("evilcorp.co.uk", None)
    assert helpers.split_host_port("192.168.0.1") == (ipaddress.ip_address("192.168.0.1"), None)
    assert helpers.split_host_port("192.168.0.1:80") == (ipaddress.ip_address("192.168.0.1"), 80)
    assert helpers.split_host_port("[e]:80") == ("e", 80)
    assert helpers.split_host_port("d://wat:wat") == ("wat", None)
    assert helpers.split_host_port("https://[dead::beef]:8338") == (ipaddress.ip_address("dead::beef"), 8338)
    assert helpers.split_host_port("[dead::beef]") == (ipaddress.ip_address("dead::beef"), None)
    assert helpers.split_host_port("dead::beef") == (ipaddress.ip_address("dead::beef"), None)
    extracted_words = helpers.extract_words("blacklanternsecurity")
    assert "black" in extracted_words
    # assert "blacklantern" in extracted_words
    # assert "lanternsecurity" in extracted_words
    # assert "blacklanternsecurity" in extracted_words
    assert "bls" in extracted_words

    choices = ["asdf.fdsa", "asdf.1234", "4321.5678"]
    best_match = helpers.closest_match("asdf.123a", choices)
    assert best_match == "asdf.1234"
    best_matches = helpers.closest_match("asdf.123a", choices, n=2)
    assert len(best_matches) == 2
    assert best_matches[0] == "asdf.1234"
    assert best_matches[1] == "asdf.fdsa"

    ipv4_netloc = helpers.make_netloc("192.168.1.1", 80)
    assert ipv4_netloc == "192.168.1.1:80"
    assert helpers.make_netloc("192.168.1.1") == "192.168.1.1"
    assert helpers.make_netloc(ipaddress.ip_address("192.168.1.1"), None) == "192.168.1.1"
    assert helpers.make_netloc("dead::beef", "443") == "[dead::beef]:443"
    assert helpers.make_netloc(ipaddress.ip_address("dead::beef"), 443) == "[dead::beef]:443"
    assert helpers.make_netloc("dead::beef", None) == "[dead::beef]"
    assert helpers.make_netloc(ipaddress.ip_address("dead::beef"), None) == "[dead::beef]"

    assert helpers.get_file_extension("https://evilcorp.com/evilcorp.com/test/asdf.TXT") == "txt"
    assert helpers.get_file_extension("/etc/conf/test.tar.gz") == "gz"
    assert helpers.get_file_extension("/etc/passwd") == ""

    assert helpers.tagify("HttP  -_Web  Title--  ") == "http-web-title"
    tagged_event = scan.make_event("127.0.0.1", parent=scan.root_event, tags=["HttP  web -__- title  "])
    assert "http-web-title" in tagged_event.tags
    tagged_event.remove_tag("http-web-title")
    assert "http-web-title" not in tagged_event.tags
    tagged_event.add_tag("Another tag  ")
    assert "another-tag" in tagged_event.tags
    tagged_event.tags = ["Some other tag  "]
    assert isinstance(tagged_event._tags, set)
    assert "another-tag" not in tagged_event.tags
    assert "some-other-tag" in tagged_event.tags

    assert list(helpers.search_dict_by_key("asdf", {"asdf": "fdsa", 4: [{"asdf": 5}]})) == ["fdsa", 5]
    assert list(helpers.search_dict_by_key("asdf", {"wat": {"asdf": "fdsa"}})) == ["fdsa"]
    assert list(helpers.search_dict_by_key("asdf", [{"wat": {"nope": 1}}, {"wat": [{"asdf": "fdsa"}]}])) == ["fdsa"]
    assert not list(helpers.search_dict_by_key("asdf", [{"wat": {"nope": 1}}, {"wat": [{"fdsa": "asdf"}]}]))
    assert not list(helpers.search_dict_by_key("asdf", "asdf"))

    from bbot.core.helpers.regexes import url_regexes

    dict_to_search = {
        "key1": {
            "key2": [{"key3": "A url of some kind: https://www.evilcorp.com/asdf"}],
            "key4": "A url of some kind: https://www.evilcorp.com/fdsa",
        }
    }
    assert set(helpers.search_dict_values(dict_to_search, *url_regexes)) == {
        "https://www.evilcorp.com/asdf",
        "https://www.evilcorp.com/fdsa",
    }

    replaced = helpers.search_format_dict(
        {"asdf": [{"wat": {"here": "#{replaceme}!"}}, {500: True}]}, replaceme="asdf"
    )
    assert replaced["asdf"][1][500] == True
    assert replaced["asdf"][0]["wat"]["here"] == "asdf!"

    filtered_dict = helpers.filter_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}}, "api_key"
    )
    assert "api_key" in filtered_dict["modules"]["c99"]
    assert "filterme" not in filtered_dict["modules"]["c99"]
    assert "ipneighbor" not in filtered_dict["modules"]

    filtered_dict2 = helpers.filter_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}}, "c99"
    )
    assert "api_key" in filtered_dict2["modules"]["c99"]
    assert "filterme" in filtered_dict2["modules"]["c99"]
    assert "ipneighbor" not in filtered_dict2["modules"]

    filtered_dict3 = helpers.filter_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}},
        "key",
        fuzzy=True,
    )
    assert "api_key" in filtered_dict3["modules"]["c99"]
    assert "filterme" not in filtered_dict3["modules"]["c99"]
    assert "ipneighbor" not in filtered_dict3["modules"]

    filtered_dict4 = helpers.filter_dict(
        {"modules": {"secrets_db": {"api_key": "1234"}, "ipneighbor": {"secret": "test", "asdf": "1234"}}},
        "secret",
        fuzzy=True,
        exclude_keys="modules",
    )
    assert not "secrets_db" in filtered_dict4["modules"]
    assert "ipneighbor" in filtered_dict4["modules"]
    assert "secret" in filtered_dict4["modules"]["ipneighbor"]
    assert "asdf" not in filtered_dict4["modules"]["ipneighbor"]

    cleaned_dict = helpers.clean_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}}, "api_key"
    )
    assert "api_key" not in cleaned_dict["modules"]["c99"]
    assert "filterme" in cleaned_dict["modules"]["c99"]
    assert "ipneighbor" in cleaned_dict["modules"]

    cleaned_dict2 = helpers.clean_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}}, "c99"
    )
    assert "c99" not in cleaned_dict2["modules"]
    assert "ipneighbor" in cleaned_dict2["modules"]

    cleaned_dict3 = helpers.clean_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}},
        "key",
        fuzzy=True,
    )
    assert "api_key" not in cleaned_dict3["modules"]["c99"]
    assert "filterme" in cleaned_dict3["modules"]["c99"]
    assert "ipneighbor" in cleaned_dict3["modules"]

    cleaned_dict4 = helpers.clean_dict(
        {"modules": {"secrets_db": {"api_key": "1234"}, "ipneighbor": {"secret": "test", "asdf": "1234"}}},
        "secret",
        fuzzy=True,
        exclude_keys="modules",
    )
    assert "secrets_db" in cleaned_dict4["modules"]
    assert "ipneighbor" in cleaned_dict4["modules"]
    assert "secret" not in cleaned_dict4["modules"]["ipneighbor"]
    assert "asdf" in cleaned_dict4["modules"]["ipneighbor"]

    assert helpers.split_list([1, 2, 3, 4, 5]) == [[1, 2], [3, 4, 5]]
    assert list(helpers.grouper("ABCDEFG", 3)) == [["A", "B", "C"], ["D", "E", "F"], ["G"]]

    assert len(helpers.rand_string(3)) == 3
    assert len(helpers.rand_string(1)) == 1
    assert len(helpers.rand_string(0)) == 0
    assert type(helpers.rand_string(0)) == str

    test_file = Path(scan.config["home"]) / "testfile.asdf"
    test_file.touch()

    assert test_file.is_file()
    backup = helpers.backup_file(test_file)
    assert backup.name == "testfile.1.asdf"
    assert not test_file.exists()
    assert backup.is_file()
    test_file.touch()
    backup2 = helpers.backup_file(test_file)
    assert backup2.name == "testfile.1.asdf"
    assert not test_file.exists()
    assert backup2.is_file()
    older_backup = Path(scan.config["home"]) / "testfile.2.asdf"
    assert older_backup.is_file()
    older_backup.unlink()
    backup.unlink()

    with open(test_file, "w") as f:
        f.write("asdf\nfdsa")

    assert "asdf" in helpers.str_or_file(str(test_file))
    assert "nope" in helpers.str_or_file("nope")
    assert tuple(helpers.chain_lists([str(test_file), "nope"], try_files=True)) == ("asdf", "fdsa", "nope")
    assert tuple(helpers.chain_lists("one, two", try_files=True)) == ("one", "two")
    assert tuple(helpers.chain_lists("one, two three ,four five")) == ("one", "two", "three", "four", "five")
    assert test_file.is_file()

    with pytest.raises(DirectoryCreationError, match="Failed to create.*"):
        helpers.mkdir(test_file)

    helpers.delete_file(test_file)
    assert not test_file.exists()

    timedelta = datetime.timedelta(hours=1, minutes=2, seconds=3)
    assert helpers.human_timedelta(timedelta) == "1 hour, 2 minutes, 3 seconds"
    timedelta = datetime.timedelta(hours=3, seconds=1)
    assert helpers.human_timedelta(timedelta) == "3 hours, 1 second"
    timedelta = datetime.timedelta(seconds=2)
    assert helpers.human_timedelta(timedelta) == "2 seconds"

    ### VALIDATORS ###
    # hosts
    assert helpers.validators.validate_host(" evilCorp.COM.") == "evilcorp.com"
    assert helpers.validators.validate_host("LOCALHOST ") == "localhost"
    assert helpers.validators.validate_host(" 192.168.1.1") == "192.168.1.1"
    assert helpers.validators.validate_host(" Dead::c0dE ") == "dead::c0de"
    assert helpers.validators.soft_validate(" evilCorp.COM", "host") == True
    assert helpers.validators.soft_validate("!@#$", "host") == False
    with pytest.raises(ValueError):
        assert helpers.validators.validate_host("!@#$")
    # ports
    assert helpers.validators.validate_port(666) == 666
    assert helpers.validators.validate_port(666666) == 65535
    assert helpers.validators.soft_validate(666, "port") == True
    assert helpers.validators.soft_validate("!@#$", "port") == False
    with pytest.raises(ValueError):
        helpers.validators.validate_port("asdf")
    # top tcp ports
    top_tcp_ports = helpers.top_tcp_ports(100)
    assert len(top_tcp_ports) == 100
    assert len(set(top_tcp_ports)) == 100
    top_tcp_ports = helpers.top_tcp_ports(800000)
    assert top_tcp_ports[:10] == [80, 23, 443, 21, 22, 25, 3389, 110, 445, 139]
    assert top_tcp_ports[-10:] == [65526, 65527, 65528, 65529, 65530, 65531, 65532, 65533, 65534, 65535]
    assert len(top_tcp_ports) == 65535
    assert len(set(top_tcp_ports)) == 65535
    assert all([isinstance(i, int) for i in top_tcp_ports])
    top_tcp_ports = helpers.top_tcp_ports(10, as_string=True)
    assert top_tcp_ports == "80,23,443,21,22,25,3389,110,445,139"
    # urls
    assert helpers.validators.validate_url(" httP://evilcorP.com/asdf?a=b&c=d#e") == "http://evilcorp.com/asdf"
    assert (
        helpers.validators.validate_url_parsed(" httP://evilcorP.com/asdf?a=b&c=d#e").geturl()
        == "http://evilcorp.com/asdf"
    )
    assert helpers.validators.soft_validate(" httP://evilcorP.com/asdf?a=b&c=d#e", "url") == True
    assert helpers.validators.soft_validate("!@#$", "url") == False
    with pytest.raises(ValueError):
        helpers.validators.validate_url("!@#$")
    # severities
    assert helpers.validators.validate_severity(" iNfo") == "INFO"
    assert helpers.validators.soft_validate(" iNfo", "severity") == True
    assert helpers.validators.soft_validate("NOPE", "severity") == False
    with pytest.raises(ValueError):
        helpers.validators.validate_severity("NOPE")
    # emails
    assert helpers.validators.validate_email(" bOb@eViLcorp.COM") == "bob@evilcorp.com"
    assert helpers.validators.soft_validate(" bOb@eViLcorp.COM", "email") == True
    assert helpers.validators.soft_validate("!@#$", "email") == False
    with pytest.raises(ValueError):
        helpers.validators.validate_email("!@#$")

    assert type(helpers.make_date()) == str

    # string formatter
    s = "asdf {unused} {used}"
    assert helpers.safe_format(s, used="fdsa") == "asdf {unused} fdsa"

    # punycode
    assert helpers.smart_encode_punycode("ドメイン.テスト") == "xn--eckwd4c7c.xn--zckzah"
    assert helpers.smart_decode_punycode("xn--eckwd4c7c.xn--zckzah") == "ドメイン.テスト"
    assert helpers.smart_encode_punycode("evilcorp.com") == "evilcorp.com"
    assert helpers.smart_decode_punycode("evilcorp.com") == "evilcorp.com"
    assert helpers.smart_encode_punycode("bob_smith@ドメイン.テスト") == "bob_smith@xn--eckwd4c7c.xn--zckzah"
    assert helpers.smart_decode_punycode("bob_smith@xn--eckwd4c7c.xn--zckzah") == "bob_smith@ドメイン.テスト"
    assert helpers.smart_encode_punycode("ドメイン.テスト:80") == "xn--eckwd4c7c.xn--zckzah:80"
    assert helpers.smart_decode_punycode("xn--eckwd4c7c.xn--zckzah:80") == "ドメイン.テスト:80"

    assert await helpers.re.recursive_decode("Hello%20world%21") == "Hello world!"
    assert (
        await helpers.re.recursive_decode("Hello%20%5Cu041f%5Cu0440%5Cu0438%5Cu0432%5Cu0435%5Cu0442") == "Hello Привет"
    )
    assert (
        await helpers.re.recursive_decode("%5Cu0020%5Cu041f%5Cu0440%5Cu0438%5Cu0432%5Cu0435%5Cu0442%5Cu0021")
        == " Привет!"
    )
    assert await helpers.re.recursive_decode("Hello%2520world%2521") == "Hello world!"
    assert (
        await helpers.re.recursive_decode(
            "Hello%255Cu0020%255Cu041f%255Cu0440%255Cu0438%255Cu0432%255Cu0435%255Cu0442"
        )
        == "Hello Привет"
    )
    assert (
        await helpers.re.recursive_decode(
            "%255Cu0020%255Cu041f%255Cu0440%255Cu0438%255Cu0432%255Cu0435%255Cu0442%255Cu0021"
        )
        == " Привет!"
    )
    assert (
        await helpers.re.recursive_decode(r"Hello\\nWorld\\\tGreetings\\\\nMore\nText")
        == "Hello\nWorld\tGreetings\nMore\nText"
    )

    ### CACHE ###
    helpers.cache_put("string", "wat")
    helpers.cache_put("binary", b"wat")
    assert helpers.cache_get("string") == "wat"
    assert helpers.cache_get("binary") == "wat"
    assert helpers.cache_get("binary", text=False) == b"wat"
    cache_filename = helpers.cache_filename("string")
    (m, i, d, n, u, g, sz, atime, mtime, ctime) = os.stat(str(cache_filename))
    # change modified time to be 10 days in the past
    os.utime(str(cache_filename), times=(atime, mtime - (3600 * 24 * 10)))
    assert helpers.cache_get("string", cache_hrs=24 * 7) is None
    assert helpers.cache_get("string", cache_hrs=24 * 14) == "wat"

    test_file = Path(scan.config["home"]) / "testfile.asdf"
    with open(test_file, "w") as f:
        for i in range(100):
            f.write(f"{i}\n")
    assert len(list(open(test_file).readlines())) == 100
    assert (await helpers.wordlist(test_file)).is_file()
    truncated_file = await helpers.wordlist(test_file, lines=10)
    assert truncated_file.is_file()
    assert len(list(open(truncated_file).readlines())) == 10
    with pytest.raises(WordlistError):
        await helpers.wordlist("/tmp/a9pseoysadf/asdkgjaosidf")
    test_file.unlink()

    # filename truncation
    super_long_filename = "/tmp/" + ("a" * 1024) + ".txt"
    with pytest.raises(OSError):
        with open(super_long_filename, "w") as f:
            f.write("wat")
    truncated_filename = helpers.truncate_filename(super_long_filename)
    with open(truncated_filename, "w") as f:
        f.write("wat")
    truncated_filename.unlink()

    # misc DNS helpers
    assert helpers.is_ptr("wsc-11-22-33-44-wat.evilcorp.com") == True
    assert helpers.is_ptr("wsc-11-22-33-wat.evilcorp.com") == False
    assert helpers.is_ptr("11wat.evilcorp.com") == False

    ## NTLM
    testheader = "TlRMTVNTUAACAAAAHgAeADgAAAAVgorilwL+bvnVipUAAAAAAAAAAJgAmABWAAAACgBjRQAAAA9XAEkATgAtAFMANAAyAE4ATwBCAEQAVgBUAEsAOAACAB4AVwBJAE4ALQBTADQAMgBOAE8AQgBEAFYAVABLADgAAQAeAFcASQBOAC0AUwA0ADIATgBPAEIARABWAFQASwA4AAQAHgBXAEkATgAtAFMANAAyAE4ATwBCAEQAVgBUAEsAOAADAB4AVwBJAE4ALQBTADQAMgBOAE8AQgBEAFYAVABLADgABwAIAHUwOZlfoNgBAAAAAA=="
    decoded = helpers.ntlm.ntlmdecode(testheader)
    assert decoded["NetBIOS_Domain_Name"] == "WIN-S42NOBDVTK8"
    assert decoded["NetBIOS_Computer_Name"] == "WIN-S42NOBDVTK8"
    assert decoded["DNS_Domain_name"] == "WIN-S42NOBDVTK8"
    assert decoded["FQDN"] == "WIN-S42NOBDVTK8"
    assert decoded["Timestamp"] == b"u09\x99_\xa0\xd8\x01"
    with pytest.raises(NTLMError):
        helpers.ntlm.ntlmdecode("asdf")

    test_filesize = bbot_test_dir / "test_filesize"
    test_filesize.touch()
    assert test_filesize.is_file()
    assert helpers.filesize(test_filesize) == 0
    assert helpers.filesize(bbot_test_dir / "glkasjdlgksadlkfsdf") == 0

    # memory stuff
    int(helpers.memory_status().available)
    int(helpers.swap_status().total)

    assert helpers.bytes_to_human(459819198709) == "428.24GB"
    assert helpers.human_to_bytes("428.24GB") == 459819198709

    # ordinals
    assert helpers.integer_to_ordinal(1) == "1st"
    assert helpers.integer_to_ordinal(2) == "2nd"
    assert helpers.integer_to_ordinal(3) == "3rd"
    assert helpers.integer_to_ordinal(4) == "4th"
    assert helpers.integer_to_ordinal(11) == "11th"
    assert helpers.integer_to_ordinal(12) == "12th"
    assert helpers.integer_to_ordinal(13) == "13th"
    assert helpers.integer_to_ordinal(21) == "21st"
    assert helpers.integer_to_ordinal(22) == "22nd"
    assert helpers.integer_to_ordinal(23) == "23rd"
    assert helpers.integer_to_ordinal(101) == "101st"
    assert helpers.integer_to_ordinal(111) == "111th"
    assert helpers.integer_to_ordinal(112) == "112th"
    assert helpers.integer_to_ordinal(113) == "113th"
    assert helpers.integer_to_ordinal(0) == "0th"

    await scan._cleanup()

    scan1 = bbot_scanner(modules="ipneighbor")
    await scan1.load_modules()
    assert int(helpers.get_size(scan1.modules["ipneighbor"])) > 0

    await scan1._cleanup()

    # weighted shuffle (used for module queues)
    items = ["a", "b", "c", "d", "e"]
    first_frequencies = {i: 0 for i in items}
    weights = [1, 2, 3, 4, 5]
    for i in range(10000):
        shuffled = helpers.weighted_shuffle(items, weights)
        first = shuffled[0]
        first_frequencies[first] += 1
    assert (
        first_frequencies["a"]
        < first_frequencies["b"]
        < first_frequencies["c"]
        < first_frequencies["d"]
        < first_frequencies["e"]
    )

    # error handling helpers
    test_ran = False
    try:
        try:
            raise KeyboardInterrupt("asdf")
        except KeyboardInterrupt:
            raise ValueError("asdf")
    except Exception as e:
        assert len(helpers.get_exception_chain(e)) == 2
        assert len([_ for _ in helpers.get_exception_chain(e) if isinstance(_, KeyboardInterrupt)]) == 1
        assert len([_ for _ in helpers.get_exception_chain(e) if isinstance(_, ValueError)]) == 1
        assert helpers.in_exception_chain(e, (KeyboardInterrupt, asyncio.CancelledError)) == True
        assert helpers.in_exception_chain(e, (TypeError, OSError)) == False
        test_ran = True
    assert test_ran
    test_ran = False
    try:
        try:
            raise AttributeError("asdf")
        except AttributeError:
            raise ValueError("asdf")
    except Exception as e:
        assert len(helpers.get_exception_chain(e)) == 2
        assert len([_ for _ in helpers.get_exception_chain(e) if isinstance(_, AttributeError)]) == 1
        assert len([_ for _ in helpers.get_exception_chain(e) if isinstance(_, ValueError)]) == 1
        assert helpers.in_exception_chain(e, (KeyboardInterrupt, asyncio.CancelledError)) == False
        assert helpers.in_exception_chain(e, (KeyboardInterrupt, AttributeError)) == True
        assert helpers.in_exception_chain(e, (AttributeError,)) == True
        test_ran = True
    assert test_ran


@pytest.mark.asyncio
async def test_word_cloud(helpers, bbot_scanner):
    number_mutations = helpers.word_cloud.get_number_mutations("base2_p013", n=5, padding=2)
    assert "base0_p013" in number_mutations
    assert "base7_p013" in number_mutations
    assert "base8_p013" not in number_mutations
    assert "base2_p008" in number_mutations
    assert "base2_p007" not in number_mutations
    assert "base2_p018" in number_mutations
    assert "base2_p0134" in number_mutations
    assert "base2_p0135" not in number_mutations

    permutations = helpers.word_cloud.mutations("_base", numbers=1)
    assert ("_base", "dev") in permutations
    assert ("dev", "_base") in permutations

    # saving and loading
    scan1 = bbot_scanner("127.0.0.1")
    word_cloud = scan1.helpers.word_cloud
    word_cloud.add_word("lantern")
    word_cloud.add_word("black")
    word_cloud.add_word("black")
    word_cloud.save()
    with open(word_cloud.default_filename) as f:
        word_cloud_content = [l.rstrip() for l in f.read().splitlines()]
    assert len(word_cloud_content) == 2
    assert "2\tblack" in word_cloud_content
    assert "1\tlantern" in word_cloud_content
    word_cloud.save(limit=1)
    with open(word_cloud.default_filename) as f:
        word_cloud_content = [l.rstrip() for l in f.read().splitlines()]
    assert len(word_cloud_content) == 1
    assert "2\tblack" in word_cloud_content
    assert "1\tlantern" not in word_cloud_content
    word_cloud.clear()
    with open(word_cloud.default_filename, "w") as f:
        f.write("plumbus\nrumbus")
    word_cloud.load()
    assert word_cloud["plumbus"] == 1
    assert word_cloud["rumbus"] == 1

    # mutators
    from bbot.core.helpers.wordcloud import DNSMutator

    m = DNSMutator()
    m.add_word("blacklantern-security237")
    mutations = set(m)
    assert mutations == {
        (None,),
        (None, "237"),
        (None, "-security237"),
        (None, "lanternsecurity237"),
        (None, "lantern-security237"),
        ("blacklantern-", None),
        ("blacklantern", None, "237"),
        ("blacklantern-", None, "237"),
        ("black", None, "security237"),
        ("black", None, "-security237"),
    }

    m = DNSMutator()
    m.add_word("blacklantern-security")
    m.add_word("sec")
    m.add_word("sec2")
    m.add_word("black2")
    mutations = sorted(m.mutations("whitebasket"))
    assert mutations == sorted(
        [
            "basket",
            "basket-security",
            "basket2",
            "basketlantern-security",
            "basketlanternsecurity",
            "blackbasket-security",
            "blackbasketsecurity",
            "blacklantern-basket",
            "blacklantern-white",
            "blacklantern-whitebasket",
            "blacklanternbasket",
            "blacklanternwhite",
            "blacklanternwhitebasket",
            "blackwhite-security",
            "blackwhitebasket-security",
            "blackwhitebasketsecurity",
            "blackwhitesecurity",
            "white",
            "white-security",
            "white2",
            "whitebasket",
            "whitebasket-security",
            "whitebasket2",
            "whitebasketlantern-security",
            "whitebasketlanternsecurity",
            "whitelantern-security",
            "whitelanternsecurity",
        ]
    )
    top_mutations = sorted(m.top_mutations().items(), key=lambda x: x[-1], reverse=True)
    assert top_mutations[:2] == [((None,), 4), ((None, "2"), 2)]

    await scan1._cleanup()


def test_names(helpers):
    assert helpers.names == sorted(helpers.names)
    assert helpers.adjectives == sorted(helpers.adjectives)


@pytest.mark.asyncio
async def test_ratelimiter(helpers):
    from bbot.core.helpers.ratelimiter import RateLimiter

    results = []

    async def web_request(r):
        async with r:
            await asyncio.sleep(0.12345)
            results.append(None)

    # allow 10 requests per second
    r = RateLimiter(10, "Test")
    tasks = []
    # start 500 requests
    for i in range(500):
        tasks.append(asyncio.create_task(web_request(r)))
    # sleep for 5 seconds
    await asyncio.sleep(5)
    await helpers.cancel_tasks(tasks)
    # 5 seconds * 10 requests per second == 50
    assert 45 <= len(results) <= 55


def test_sync_to_async():
    from bbot.core.helpers.async_helpers import async_to_sync_gen

    # async to sync generator converter
    async def async_gen():
        for i in range(5):
            await asyncio.sleep(0.1)
            yield i

    sync_gen = async_to_sync_gen(async_gen())

    l = []
    while 1:
        try:
            l.append(next(sync_gen))
        except StopIteration:
            break
    assert l == [0, 1, 2, 3, 4]


@pytest.mark.asyncio
async def test_async_helpers():
    import random
    from bbot.core.helpers.misc import as_completed

    async def do_stuff(r):
        await asyncio.sleep(r)
        return r

    random_ints = [random.random() for _ in range(1000)]
    tasks = [do_stuff(r) for r in random_ints]
    results = set()
    async for t in as_completed(tasks):
        results.add(await t)
    assert len(results) == 1000
    assert sorted(random_ints) == sorted(results)


def test_portparse(helpers):
    assert helpers.parse_port_string("80,443,22") == [80, 443, 22]
    assert helpers.parse_port_string(80) == [80]

    assert helpers.parse_port_string("80,443,22,1000-1002") == [80, 443, 22, 1000, 1001, 1002]

    with pytest.raises(ValueError) as e:
        helpers.parse_port_string("80,443,22,70000")
    assert str(e.value) == "Invalid port: 70000"

    with pytest.raises(ValueError) as e:
        helpers.parse_port_string("80,443,22,1000-70000")
    assert str(e.value) == "Invalid port range: 1000-70000"

    with pytest.raises(ValueError) as e:
        helpers.parse_port_string("80,443,22,1000-1001-1002")
    assert str(e.value) == "Invalid port or port range: 1000-1001-1002"

    with pytest.raises(ValueError) as e:
        helpers.parse_port_string("80,443,22,1002-1000")
    assert str(e.value) == "Invalid port range: 1002-1000"

    with pytest.raises(ValueError) as e:
        helpers.parse_port_string("80,443,22,foo")
    assert str(e.value) == "Invalid port or port range: foo"


# test chain_lists helper


def test_liststring_valid_strings(helpers):
    assert helpers.chain_lists("hello,world,bbot") == ["hello", "world", "bbot"]


def test_liststring_invalid_string(helpers):
    with pytest.raises(ValueError) as e:
        helpers.chain_lists("hello,world,\x01", validate=True)
    assert str(e.value) == "Invalid character in string: \x01"


def test_liststring_singleitem(helpers):
    assert helpers.chain_lists("hello") == ["hello"]


def test_liststring_invalidfnchars(helpers):
    with pytest.raises(ValueError) as e:
        helpers.chain_lists("hello,world,bbot|test", validate=True)
    assert str(e.value) == "Invalid character in string: bbot|test"


# test parameter validation
@pytest.mark.asyncio
async def test_parameter_validation(helpers):

    getparam_valid_params = {
        "name",
        "age",
        "valid_name",
        "valid-name",
        "session_token",
        "user.id",
        "user-name",
        "client.id",
        "auth-token",
        "access_token",
        "abcd",
        "jqueryget",
        "<script>",
    }
    getparam_invalid_params = {
        "invalid,name",
        "###$$$",
        "this_parameter_name_is_seriously_way_too_long_to_be_practical_but_hey_look_its_still_technically_valid_wow",
        "parens()",
        "cookie$name",
    }

    getparam_params = getparam_valid_params | getparam_invalid_params
    for p in getparam_params:
        if helpers.validate_parameter(p, "getparam"):
            assert p in getparam_valid_params and p not in getparam_invalid_params
        else:
            assert p in getparam_invalid_params and not p in getparam_valid_params

    header_valid_params = {
        "name",
        "age",
        "valid_name",
        "valid-name",
        "session_token",
        "user-name",
        "auth-token",
        "access_token",
        "abcd",
        "jqueryget",
    }
    header_invalid_params = {
        "invalid,name",
        "<script>",
        "this_parameter_name_is_seriously_way_too_long_to_be_practical_but_hey_look_its_still_technically_valid_wow",
        "parens()",
        "cookie$name",
        "carrot^",
        "###$$$",
        "user.id",
        "client.id",
    }

    header_params = header_valid_params | header_invalid_params
    for p in header_params:
        if helpers.validate_parameter(p, "header"):
            assert p in header_valid_params and p not in header_invalid_params
        else:
            assert p in header_invalid_params and not p in header_valid_params

    cookie_valid_params = {
        "name",
        "age",
        "valid_name",
        "valid-name",
        "session_token",
        "user-name",
        "auth-token",
        "access_token",
        "user.id",
        "client.id",
        "abcd",
        "jqueryget",
        "###$$$",
        "cookie$name",
    }
    cookie_invalid_params = {
        "invalid,name",
        "<script>",
        "parens()",
        "this_parameter_name_is_seriously_way_too_long_to_be_practical_but_hey_look_its_still_technically_valid_wow",
    }

    cookie_params = cookie_valid_params | cookie_invalid_params
    for p in cookie_params:
        if helpers.validate_parameter(p, "cookie"):
            assert p in cookie_valid_params and p not in cookie_invalid_params
        else:
            assert p in cookie_invalid_params and not p in cookie_valid_params



import json

from .base import ModuleTestBase
from bbot.core.event.base import event_from_json


class TestJSON(ModuleTestBase):
    def check(self, module_test, events):
        dns_data = "blacklanternsecurity.com"
        context_data = f"Scan {module_test.scan.name} seeded with DNS_NAME: blacklanternsecurity.com"

        scan_event = [e for e in events if e.type == "SCAN"][0]
        dns_event = [e for e in events if e.type == "DNS_NAME"][0]

        # json events
        txt_file = module_test.scan.home / "output.json"
        lines = list(module_test.scan.helpers.read_file(txt_file))
        assert lines
        json_events = [json.loads(line) for line in lines]
        scan_json = [e for e in json_events if e["type"] == "SCAN"]
        dns_json = [e for e in json_events if e["type"] == "DNS_NAME"]
        assert len(scan_json) == 2
        assert len(dns_json) == 1
        dns_json = dns_json[0]
        scan = scan_json[0]
        assert scan["data"]["name"] == module_test.scan.name
        assert scan["data"]["id"] == module_test.scan.id
        assert scan["id"] == module_test.scan.id
        assert scan["uuid"] == str(module_test.scan.root_event.uuid)
        assert scan["parent_uuid"] == str(module_test.scan.root_event.uuid)
        assert scan["data"]["target"]["seeds"] == ["blacklanternsecurity.com"]
        assert scan["data"]["target"]["whitelist"] == ["blacklanternsecurity.com"]
        assert dns_json["data"] == dns_data
        assert dns_json["id"] == str(dns_event.id)
        assert dns_json["uuid"] == str(dns_event.uuid)
        assert dns_json["parent_uuid"] == str(module_test.scan.root_event.uuid)
        assert dns_json["discovery_context"] == context_data
        assert dns_json["discovery_path"] == [context_data]
        assert dns_json["parent_chain"] == [dns_json["uuid"]]

        # event objects reconstructed from json
        scan_reconstructed = event_from_json(scan_json[0])
        dns_reconstructed = event_from_json(dns_json)
        assert scan_reconstructed.data["name"] == module_test.scan.name
        assert scan_reconstructed.data["id"] == module_test.scan.id
        assert scan_reconstructed.uuid == scan_event.uuid
        assert scan_reconstructed.parent_uuid == scan_event.uuid
        assert scan_reconstructed.data["target"]["seeds"] == ["blacklanternsecurity.com"]
        assert scan_reconstructed.data["target"]["whitelist"] == ["blacklanternsecurity.com"]
        assert dns_reconstructed.data == dns_data
        assert dns_reconstructed.uuid == dns_event.uuid
        assert dns_reconstructed.parent_uuid == module_test.scan.root_event.uuid
        assert dns_reconstructed.discovery_context == context_data
        assert dns_reconstructed.discovery_path == [context_data]
        assert dns_reconstructed.parent_chain == [dns_json["uuid"]]


class TestJSONSIEMFriendly(ModuleTestBase):
    modules_overrides = ["json"]
    config_overrides = {"modules": {"json": {"siem_friendly": True}}}

    def check(self, module_test, events):
        txt_file = module_test.scan.home / "output.json"
        lines = list(module_test.scan.helpers.read_file(txt_file))
        passed = False
        for line in lines:
            e = json.loads(line)
            if e["data"] == {"DNS_NAME": "blacklanternsecurity.com"}:
                passed = True
        assert passed



from bbot.modules.templates.shodan import shodan


class shodan_dns(shodan):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query Shodan for subdomains",
        "created_date": "2022-07-03",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "Shodan API key"}

    base_url = "https://api.shodan.io"

    async def handle_event(self, event):
        await self.handle_event_paginated(event)

    def make_url(self, query):
        return f"{self.base_url}/dns/domain/{self.helpers.quote(query)}?key={{api_key}}&page={{page}}"

    def parse_results(self, json, query):
        return [f"{sub}.{query}" for sub in json.get("subdomains", [])]



import io
import re
import uuid
import json
import base64
import logging
import tarfile
import datetime
import ipaddress
import traceback

from copy import copy
from pathlib import Path
from typing import Optional
from contextlib import suppress
from radixtarget import RadixTarget
from urllib.parse import urljoin, parse_qs
from pydantic import BaseModel, field_validator


from .helpers import *
from bbot.errors import *
from bbot.core.helpers import (
    extract_words,
    is_domain,
    is_subdomain,
    is_ip,
    is_ip_type,
    is_ptr,
    is_uri,
    url_depth,
    domain_stem,
    make_netloc,
    make_ip_type,
    recursive_decode,
    sha1,
    smart_decode,
    split_host_port,
    tagify,
    validators,
    get_file_extension,
)


log = logging.getLogger("bbot.core.event")


class BaseEvent:
    """
    Represents a piece of data discovered during a BBOT scan.

    An Event contains various attributes that provide metadata about the discovered data.
    The attributes assist in understanding the context of the Event and facilitate further
    filtering and querying. Events are integral in the construction of visual graphs and
    are the cornerstone of data exchange between BBOT modules.

    You can inherit from this class when creating a new event type. However, it's not always
    necessary. You only need to subclass if you want to layer additional functionality on
    top of the base class.

    Attributes:
        type (str): Specifies the type of the event, e.g., `IP_ADDRESS`, `DNS_NAME`.
        id (str): An identifier for the event (event type + sha1 hash of data). NOT universally unique.
        uuid (UUID): A universally unique identifier for the event.
        data (str or dict): The main data for the event, e.g., a URL or IP address.
        data_graph (str): Representation of `self.data` for graph nodes (e.g. Neo4j).
        data_human (str): Representation of `self.data` for human output.
        data_id (str): Representation of `self.data` used to calculate the event's ID (and ultimately its hash, which is used for deduplication)
        data_json (str): Representation of `self.data` to be used in JSON serialization.
        host (str, IPvXAddress, or IPvXNetwork): The associated IP address or hostname for the event
        host_stem (str): An abbreviated representation of hostname that removes the TLD, e.g. "www.evilcorp". Used by the word cloud.
        port (int or None): The port associated with the event, if applicable, else None.
        words (set): A list of relevant keywords extracted from the event. Used by the word cloud.
        scope_distance (int): Indicates how many hops the event is from the main scope; 0 means in-scope.
        web_spider_distance (int): The spider distance from the web root, specific to web crawling.
        scan (Scanner): The scan object that generated the event.
        timestamp (datetime.datetime): The time at which the data was discovered.
        resolved_hosts (list of str): List of hosts to which the event data resolves, applicable for URLs and DNS names.
        parent (BaseEvent): The parent event that led to the discovery of this event.
        parent_id (str): The `id` attribute of the parent event.
        parent_uuid (str): The `uuid` attribute of the parent event.
        tags (set of str): Descriptive tags for the event, e.g., `mx-record`, `in-scope`.
        module (BaseModule): The module that discovered the event.
        module_sequence (str): The sequence of modules that participated in the discovery.

    Examples:
        ```json
        {
            "type": "URL",
            "id": "URL:017ec8e5dc158c0fd46f07169f8577fb4b45e89a",
            "data": "http://www.blacklanternsecurity.com/",
            "web_spider_distance": 0,
            "scope_distance": 0,
            "scan": "SCAN:4d786912dbc97be199da13074699c318e2067a7f",
            "timestamp": 1688526222.723366,
            "resolved_hosts": ["185.199.108.153"],
            "parent": "OPEN_TCP_PORT:cf7e6a937b161217eaed99f0c566eae045d094c7",
            "tags": ["in-scope", "distance-0", "dir", "ip-185-199-108-153", "status-301", "http-title-301-moved-permanently"],
            "module": "httpx",
            "module_sequence": "httpx"
        }
        ```
    """

    # Always emit this event type even if it's not in scope
    _always_emit = False
    # Always emit events with these tags even if they're not in scope
    _always_emit_tags = ["affiliate", "target"]
    # Bypass scope checking and dns resolution, distribute immediately to modules
    # This is useful for "end-of-line" events like FINDING and VULNERABILITY
    _quick_emit = False
    # Whether this event has been retroactively marked as part of an important discovery chain
    _graph_important = False
    # Disables certain data validations
    _dummy = False
    # Data validation, if data is a dictionary
    _data_validator = None
    # Whether to increment scope distance if the child and parent hosts are the same
    _scope_distance_increment_same_host = False
    # Don't allow duplicates to occur within a parent chain
    # In other words, don't emit the event if the same one already exists in its discovery context
    _suppress_chain_dupes = False

    def __init__(
        self,
        data,
        event_type,
        parent=None,
        context=None,
        module=None,
        scan=None,
        scans=None,
        tags=None,
        confidence=100,
        timestamp=None,
        _dummy=False,
        _internal=None,
    ):
        """
        Initializes an Event object with the given parameters.

        In most cases, you should use `make_event()` instead of instantiating this class directly.
        `make_event()` is much friendlier, and can auto-detect the event type for you.

        Attributes:
            data (str, dict): The primary data for the event.
            event_type (str, optional): Type of the event, e.g., 'IP_ADDRESS'.
            parent (BaseEvent, optional): Parent event that led to this event's discovery. Defaults to None.
            module (str, optional): Module that discovered the event. Defaults to None.
            scan (Scan, optional): BBOT Scan object. Required unless _dummy is True. Defaults to None.
            scans (list of Scan, optional): BBOT Scan objects, used primarily when unserializing an Event from the database. Defaults to None.
            tags (list of str, optional): Descriptive tags for the event. Defaults to None.
            confidence (int, optional): Confidence level for the event, on a scale of 1-100. Defaults to 100.
            timestamp (datetime, optional): Time of event discovery. Defaults to current UTC time.
            _dummy (bool, optional): If True, disables certain data validations. Defaults to False.
            _internal (Any, optional): If specified, makes the event internal. Defaults to None.

        Raises:
            ValidationError: If either `scan` or `parent` are not specified and `_dummy` is False.
        """
        self._uuid = uuid.uuid4()
        self._id = None
        self._hash = None
        self._data = None
        self.__host = None
        self._tags = set()
        self._port = None
        self._omit = False
        self.__words = None
        self._parent = None
        self._priority = None
        self._parent_id = None
        self._parent_uuid = None
        self._host_original = None
        self._scope_distance = None
        self._module_priority = None
        self._resolved_hosts = set()
        self.dns_children = dict()
        self.raw_dns_records = dict()
        self._discovery_context = ""
        self._discovery_context_regex = re.compile(r"\{(?:event|module)[^}]*\}")
        self.web_spider_distance = 0

        # for creating one-off events without enforcing parent requirement
        self._dummy = _dummy
        self.module = module
        self._type = event_type

        # keep track of whether this event has been recorded by the scan
        self._stats_recorded = False

        if timestamp is not None:
            self.timestamp = timestamp
        else:
            try:
                self.timestamp = datetime.datetime.now(datetime.UTC)
            except AttributeError:
                self.timestamp = datetime.datetime.utcnow()

        self.confidence = int(confidence)
        self._internal = False

        # self.scan holds the instantiated scan object (for helpers, etc.)
        self.scan = scan
        if (not self.scan) and (not self._dummy):
            raise ValidationError(f"Must specify scan")
        # self.scans holds a list of scan IDs from scans that encountered this event
        self.scans = []
        if scans is not None:
            self.scans = scans
        if self.scan:
            self.scans = list(set([self.scan.id] + self.scans))

        try:
            self.data = self._sanitize_data(data)
        except Exception as e:
            log.trace(traceback.format_exc())
            raise ValidationError(f'Error sanitizing event data "{data}" for type "{self.type}": {e}')

        if not self.data:
            raise ValidationError(f'Invalid event data "{data}" for type "{self.type}"')

        self.parent = parent
        if (not self.parent) and (not self._dummy):
            raise ValidationError(f"Must specify event parent")

        if tags is not None:
            for tag in tags:
                self.add_tag(tag)

        # internal events are not ingested by output modules
        if not self._dummy:
            # removed this second part because it was making certain sslcert events internal
            if _internal:  # or parent._internal:
                self.internal = True

        if not context:
            context = getattr(self.module, "default_discovery_context", "")
        if context:
            self.discovery_context = context

    @property
    def data(self):
        return self._data

    @property
    def confidence(self):
        return self._confidence

    @confidence.setter
    def confidence(self, confidence):
        self._confidence = min(100, max(1, int(confidence)))

    @property
    def cumulative_confidence(self):
        """
        Considers the confidence of parent events. This is useful for filtering out speculative/unreliable events.

        E.g. an event with a confidence of 50 whose parent is also 50 would have a cumulative confidence of 25.

        A confidence of 100 will reset the cumulative confidence to 100.
        """
        if self._confidence == 100 or self.parent is None or self.parent is self:
            return self._confidence
        return int(self._confidence * self.parent.cumulative_confidence / 100)

    @property
    def resolved_hosts(self):
        if is_ip(self.host):
            return {
                self.host,
            }
        return self._resolved_hosts

    @data.setter
    def data(self, data):
        self._hash = None
        self._data_hash = None
        self._id = None
        self.__host = None
        self._port = None
        self._data = data

    @property
    def internal(self):
        return self._internal

    @internal.setter
    def internal(self, value):
        """
        Marks the event as internal, excluding it from output but allowing normal exchange between scan modules.

        Internal events are typically speculative and may not be interesting by themselves but can lead to
        the discovery of interesting events. This method sets the `_internal` attribute to True and adds the
        "internal" tag.

        Examples of internal events include `OPEN_TCP_PORT`s from the `speculate` module,
        `IP_ADDRESS`es from the `ipneighbor` module, or out-of-scope `DNS_NAME`s that originate
        from DNS resolutions.

        The purpose of internal events is to enable speculative/explorative discovery without cluttering
        the console with irrelevant or uninteresting events.
        """
        if not value in (True, False):
            raise ValueError(f'"internal" must be boolean, not {type(value)}')
        if value == True:
            self.add_tag("internal")
        else:
            self.remove_tag("internal")
        self._internal = value

    @property
    def host(self):
        """
        An abbreviated representation of the data that allows comparison with other events.
        For host types, this is a hostname.
        This allows comparison of an email or a URL with a domain, and vice versa
            bob@evilcorp.com        --> evilcorp.com
            https://evilcorp.com    --> evilcorp.com
            evilcorp.com:80         --> evilcorp.com

        For IP_* types, this is an instantiated object representing the event's data
        E.g. for IP_ADDRESS, it could be an ipaddress.IPv4Address() or IPv6Address() object
        """
        if self.__host is None:
            self.host = self._host()
        return self.__host

    @host.setter
    def host(self, host):
        if self._host_original is None:
            self._host_original = host
        self.__host = host

    @property
    def host_original(self):
        """
        Original host data, in case it was changed due to a wildcard DNS, etc.
        """
        if self._host_original is None:
            return self.host
        return self._host_original

    @property
    def port(self):
        self.host
        if getattr(self, "parsed_url", None):
            if self.parsed_url.port is not None:
                return self.parsed_url.port
            elif self.parsed_url.scheme == "https":
                return 443
            elif self.parsed_url.scheme == "http":
                return 80
        return self._port

    @property
    def netloc(self):
        if self.host and is_ip_type(self.host, network=False):
            return make_netloc(self.host, self.port)
        return None

    @property
    def host_stem(self):
        """
        An abbreviated representation of hostname that removes the TLD
            E.g. www.evilcorp.com --> www.evilcorp
        """
        if self.host and type(self.host) == str:
            return domain_stem(self.host)
        else:
            return f"{self.host}"

    @property
    def discovery_context(self):
        return self._discovery_context

    @discovery_context.setter
    def discovery_context(self, context):
        def replace(match):
            s = match.group()
            return s.format(module=self.module, event=self)

        try:
            self._discovery_context = self._discovery_context_regex.sub(replace, context)
        except Exception as e:
            log.trace(f"Error formatting discovery context for {self}: {e} (context: '{context}')")
            self._discovery_context = context

    @property
    def discovery_path(self):
        """
        This event's full discovery context, including those of all its parents
        """
        discovery_path = []
        if self.parent is not None and self.parent is not self:
            discovery_path = self.parent.discovery_path
        return discovery_path + [self.discovery_context]

    @property
    def parent_chain(self):
        """
        This event's full discovery context, including those of all its parents
        """
        parent_chain = []
        if self.parent is not None and self.parent is not self:
            parent_chain = self.parent.parent_chain
        return parent_chain + [str(self.uuid)]

    @property
    def words(self):
        if self.__words is None:
            self.__words = set(self._words())
        return self.__words

    def _words(self):
        return set()

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, tags):
        self._tags = set()
        if isinstance(tags, str):
            tags = (tags,)
        for tag in tags:
            self.add_tag(tag)

    def add_tag(self, tag):
        self._tags.add(tagify(tag))

    def add_tags(self, tags):
        for tag in set(tags):
            self.add_tag(tag)

    def remove_tag(self, tag):
        with suppress(KeyError):
            self._tags.remove(tagify(tag))

    @property
    def always_emit(self):
        """
        If this returns True, the event will always be distributed to output modules regardless of scope distance
        """
        always_emit_tags = any(t in self.tags for t in self._always_emit_tags)
        no_host_information = not bool(self.host)
        return self._always_emit or always_emit_tags or no_host_information

    @property
    def id(self):
        """
        A uniquely identifiable hash of the event from the event type + a SHA1 of its data
        """
        if self._id is None:
            self._id = f"{self.type}:{self.data_hash.hex()}"
        return self._id

    @property
    def uuid(self):
        """
        A universally unique identifier for the event
        """
        return f"{self.type}:{self._uuid}"

    @property
    def data_hash(self):
        """
        A raw byte hash of the event's data
        """
        if self._data_hash is None:
            self._data_hash = sha1(self.data_id).digest()
        return self._data_hash

    @property
    def scope_distance(self):
        return self._scope_distance

    @scope_distance.setter
    def scope_distance(self, scope_distance):
        """
        Setter for the scope_distance attribute, ensuring it only decreases.

        The scope_distance attribute is designed to never increase; it can only be set to smaller values than
        the current one. If a larger value is provided, it is ignored. The setter also updates the event's
        tags to reflect the new scope distance.

        Parameters:
            scope_distance (int): The new scope distance to set, must be a non-negative integer.

        Note:
            The method will automatically update the relevant 'distance-' tags associated with the event.
        """
        if scope_distance < 0:
            raise ValueError(f"Invalid scope distance: {scope_distance}")
        # ensure scope distance does not increase (only allow setting to smaller values)
        if self.scope_distance is None:
            new_scope_distance = scope_distance
        else:
            new_scope_distance = min(self.scope_distance, scope_distance)
        if self._scope_distance != new_scope_distance:
            # remove old scope distance tags
            for t in list(self.tags):
                if t.startswith("distance-"):
                    self.remove_tag(t)
            if scope_distance == 0:
                self.add_tag("in-scope")
                self.remove_tag("affiliate")
            else:
                self.remove_tag("in-scope")
                self.add_tag(f"distance-{new_scope_distance}")
            self._scope_distance = new_scope_distance
            # apply recursively to parent events
            parent_scope_distance = getattr(self.parent, "scope_distance", None)
            if parent_scope_distance is not None and self.parent is not self:
                self.parent.scope_distance = new_scope_distance + 1

    @property
    def scope_description(self):
        """
        Returns a single word describing the scope of the event.

        "in-scope" if the event is in scope, "affiliate" if it's an affiliate, otherwise "distance-{scope_distance}"
        """
        if self.scope_distance == 0:
            return "in-scope"
        elif "affiliate" in self.tags:
            return "affiliate"
        return f"distance-{self.scope_distance}"

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        """
        Setter for the parent attribute, ensuring it's a valid event and updating scope distance.

        Sets the parent of the event and automatically adjusts the scope distance based on the parent event's
        scope distance. The scope distance is incremented by 1 if the host of the parent event is different
        from the current event's host.

        Parameters:
            parent (BaseEvent): The new parent event to set. Must be a valid event object.

        Note:
            If an invalid parent is provided and the event is not a dummy, a warning will be logged.
        """
        if is_event(parent):
            self._parent = parent
            hosts_are_same = (self.host and parent.host) and (self.host == parent.host)
            new_scope_distance = int(parent.scope_distance)
            if self.host and parent.scope_distance is not None:
                # only increment the scope distance if the host changes
                if self._scope_distance_increment_same_host or not hosts_are_same:
                    new_scope_distance += 1
            self.scope_distance = new_scope_distance
            # inherit certain tags
            if hosts_are_same:
                # inherit web spider distance from parent
                self.web_spider_distance = getattr(parent, "web_spider_distance", 0)
                event_has_url = getattr(self, "parsed_url", None) is not None
                for t in parent.tags:
                    if t in ("affiliate",):
                        self.add_tag(t)
                    elif t.startswith("mutation-"):
                        self.add_tag(t)
                    # only add these tags if the event has a URL
                    if event_has_url:
                        if t in ("spider-danger", "spider-max"):
                            self.add_tag(t)
        elif not self._dummy:
            log.warning(f"Tried to set invalid parent on {self}: (got: {parent})")

    @property
    def parent_id(self):
        parent_id = getattr(self.get_parent(), "id", None)
        if parent_id is not None:
            return parent_id
        return self._parent_id

    @property
    def parent_uuid(self):
        parent_uuid = getattr(self.get_parent(), "uuid", None)
        if parent_uuid is not None:
            return parent_uuid
        return self._parent_uuid

    @property
    def validators(self):
        """
        Depending on whether the scan attribute is accessible, return either a config-aware or non-config-aware validator

        This exists to prevent a chicken-and-egg scenario during the creation of certain events such as URLs,
        whose sanitization behavior is different depending on the config.

        However, thanks to this property, validation can still work in the absence of a config.
        """
        if self.scan is not None:
            return self.scan.helpers.config_aware_validators
        return validators

    def get_parent(self):
        """
        Takes into account events with the _omit flag
        """
        if getattr(self.parent, "_omit", False):
            return self.parent.get_parent()
        return self.parent

    def get_parents(self, omit=False, include_self=False):
        parents = []
        e = self
        if include_self:
            parents.append(self)
        while 1:
            if omit:
                parent = e.get_parent()
            else:
                parent = e.parent
            if parent is None:
                break
            if e == parent:
                break
            parents.append(parent)
            e = parent
        return parents

    def _host(self):
        return ""

    def _sanitize_data(self, data):
        """
        Validates and sanitizes the event's data during instantiation.

        By default, uses the '_data_load' method to pre-process the data and then applies the '_data_validator'
        to validate and create a sanitized dictionary. Raises a ValidationError if any of the validations fail.
        Subclasses can override this method to provide custom validation logic.

        Returns:
            Any: The sanitized data.

        Raises:
            ValidationError: If the data fails to validate.
        """
        data = self._data_load(data)
        if self._data_validator is not None:
            if not isinstance(data, dict):
                raise ValidationError(f"data is not of type dict: {data}")
            data = self._data_validator(**data).model_dump(exclude_none=True)
        return self.sanitize_data(data)

    def sanitize_data(self, data):
        return data

    @property
    def data_human(self):
        """
        Human representation of event.data
        """
        return self._data_human()

    def _data_human(self):
        if isinstance(self.data, (dict, list)):
            with suppress(Exception):
                return json.dumps(self.data, sort_keys=True)
        return smart_decode(self.data)

    def _data_load(self, data):
        """
        How to load the event data (JSON-decode it, etc.)
        """
        return data

    @property
    def data_id(self):
        """
        Representation of the event.data used to calculate the event's ID
        """
        return self._data_id()

    def _data_id(self):
        return self.data

    @property
    def pretty_string(self):
        """
        A human-friendly representation of the event's data. Used for graph representation.

        If the event's data is a dictionary, the function will try to return a JSON-formatted string.
        Otherwise, it will use smart_decode to convert the data into a string representation.

        Override if necessary.

        Returns:
            str: The graphical representation of the event's data.
        """
        return self._pretty_string()

    def _pretty_string(self):
        return self._data_human()

    @property
    def data_graph(self):
        """
        Representation of event.data for neo4j graph nodes
        """
        return self.pretty_string

    @property
    def data_json(self):
        """
        JSON representation of event.data
        """
        return self.data

    def __contains__(self, other):
        """
        Allows events to be compared using the "in" operator:
        E.g.:
            if some_event in other_event:
                ...
        """
        try:
            other = make_event(other, dummy=True)
        except ValidationError:
            return False
        # if hashes match
        if other == self:
            return True
        # if hosts match
        if self.host and other.host:
            if self.host == other.host:
                return True
            # hostnames and IPs
            radixtarget = RadixTarget()
            radixtarget.insert(self.host)
            return bool(radixtarget.search(other.host))
        return False

    def json(self, mode="json", siem_friendly=False):
        """
        Serializes the event object to a JSON-compatible dictionary.

        By default, it includes attributes such as 'type', 'id', 'data', 'scope_distance', and others that are present.
        Additional specific attributes can be serialized based on the mode specified.

        Parameters:
            mode (str): Specifies the data serialization mode. Default is "json". Other options include "graph", "human", and "id".
            siem_friendly (bool): Whether to format the JSON in a way that's friendly to SIEM ingestion by Elastic, Splunk, etc. This ensures the value of "data" is always the same type (a dictionary).

        Returns:
            dict: JSON-serializable dictionary representation of the event object.
        """
        j = dict()
        # type, ID, scope description
        for i in ("type", "id", "uuid", "scope_description", "netloc"):
            v = getattr(self, i, "")
            if v:
                j.update({i: str(v)})
        # event data
        data_attr = getattr(self, f"data_{mode}", None)
        if data_attr is not None:
            data = data_attr
        else:
            data = smart_decode(self.data)
        if siem_friendly:
            j["data"] = {self.type: data}
        else:
            j["data"] = data
        # host, dns children
        if self.host:
            j["host"] = str(self.host)
            j["resolved_hosts"] = sorted(str(h) for h in self.resolved_hosts)
            j["dns_children"] = {k: list(v) for k, v in self.dns_children.items()}
        if isinstance(self.port, int):
            j["port"] = self.port
        # web spider distance
        web_spider_distance = getattr(self, "web_spider_distance", None)
        if web_spider_distance is not None:
            j["web_spider_distance"] = web_spider_distance
        # scope distance
        j["scope_distance"] = self.scope_distance
        # scan
        if self.scan:
            j["scan"] = self.scan.id
        # timestamp
        j["timestamp"] = self.timestamp.isoformat()
        # parent event
        parent_id = self.parent_id
        if parent_id:
            j["parent"] = parent_id
        parent_uuid = self.parent_uuid
        if parent_uuid:
            j["parent_uuid"] = parent_uuid
        # tags
        if self.tags:
            j.update({"tags": list(self.tags)})
        # parent module
        if self.module:
            j.update({"module": str(self.module)})
        # sequence of modules that led to discovery
        if self.module_sequence:
            j.update({"module_sequence": str(self.module_sequence)})
        # discovery context
        j["discovery_context"] = self.discovery_context
        j["discovery_path"] = self.discovery_path
        j["parent_chain"] = self.parent_chain

        # normalize non-primitive python objects
        for k, v in list(j.items()):
            if k == "data":
                continue
            if type(v) not in (str, int, float, bool, list, dict, type(None)):
                try:
                    j[k] = json.dumps(v, sort_keys=True)
                except Exception:
                    j[k] = smart_decode(v)
        return j

    @staticmethod
    def from_json(j):
        """
        Convenience shortcut to create an Event object from a JSON-compatible dictionary.

        Calls the `event_from_json()` function to deserialize the event.

        Parameters:
            j (dict): The JSON-compatible dictionary containing event data.

        Returns:
            Event: The deserialized Event object.
        """
        return event_from_json(j)

    @property
    def module_sequence(self):
        """
        Get a human-friendly string that represents the sequence of modules responsible for generating this event.

        Includes the names of omitted parent events to provide a complete view of the module sequence leading to this event.

        Returns:
            str: The module sequence in human-friendly format.
        """
        module_name = getattr(self.module, "name", "")
        if getattr(self.parent, "_omit", False):
            module_name = f"{self.parent.module_sequence}->{module_name}"
        return module_name

    @property
    def module_priority(self):
        if self._module_priority is None:
            module = getattr(self, "module", None)
            self._module_priority = int(max(1, min(5, getattr(module, "priority", 3))))
        return self._module_priority

    @module_priority.setter
    def module_priority(self, priority):
        self._module_priority = int(max(1, min(5, priority)))

    @property
    def priority(self):
        if self._priority is None:
            timestamp = self.timestamp.timestamp()
            if self.parent.timestamp == self.timestamp:
                self._priority = (timestamp,)
            else:
                self._priority = getattr(self.parent, "priority", ()) + (timestamp,)

        return self._priority

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, val):
        self._type = val
        self._hash = None
        self._id = None

    @property
    def _host_size(self):
        """
        Used for sorting events by their host size, so that parent ones (e.g. IP subnets) come first
        """
        if self.host:
            if isinstance(self.host, str):
                # smaller domains should come first
                return len(self.host)
            else:
                try:
                    # bigger IP subnets should come first
                    return -self.host.num_addresses
                except AttributeError:
                    # IP addresses default to 1
                    return 1
        return 0

    def __iter__(self):
        """
        For dict(event)
        """
        yield from self.json().items()

    def __lt__(self, other):
        """
        For queue sorting
        """
        return self.priority < getattr(other, "priority", (0,))

    def __gt__(self, other):
        """
        For queue sorting
        """
        return self.priority > getattr(other, "priority", (0,))

    def __eq__(self, other):
        try:
            other = make_event(other, dummy=True)
        except ValidationError:
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.id)
        return self._hash

    def __str__(self):
        max_event_len = 80
        d = str(self.data).replace("\n", "\\n")
        return f'{self.type}("{d[:max_event_len]}{("..." if len(d) > max_event_len else "")}", module={self.module}, tags={self.tags})'

    def __repr__(self):
        return str(self)


class SCAN(BaseEvent):
    def _data_human(self):
        return f"{self.data['name']} ({self.data['id']})"

    @property
    def discovery_path(self):
        return []

    @property
    def parent_chain(self):
        return []


class FINISHED(BaseEvent):
    """
    Special signal event to indicate end of scan
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._priority = (999999999999999,)


class DefaultEvent(BaseEvent):
    def sanitize_data(self, data):
        return data


class DictEvent(BaseEvent):
    def sanitize_data(self, data):
        url = data.get("url", "")
        if url:
            self.parsed_url = self.validators.validate_url_parsed(url)
        return data

    def _data_load(self, data):
        if isinstance(data, str):
            return json.loads(data)
        return data


class DictHostEvent(DictEvent):
    def _host(self):
        if isinstance(self.data, dict) and "host" in self.data:
            return make_ip_type(self.data["host"])
        else:
            parsed = getattr(self, "parsed_url", None)
            if parsed is not None:
                return make_ip_type(parsed.hostname)


class ClosestHostEvent(DictHostEvent):
    # if a host/path/url isn't specified, this event type grabs it from the closest parent
    # inherited by FINDING and VULNERABILITY
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.host:
            for parent in self.get_parents(include_self=True):
                # inherit closest URL
                if not "url" in self.data:
                    parent_url = getattr(parent, "parsed_url", None)
                    if parent_url is not None:
                        self.data["url"] = parent_url.geturl()
                # inherit closest path
                if not "path" in self.data and isinstance(parent.data, dict):
                    parent_path = parent.data.get("path", None)
                    if parent_path is not None:
                        self.data["path"] = parent_path
                # inherit closest host
                if parent.host:
                    self.data["host"] = str(parent.host)
                    break
        # die if we still haven't found a host
        if not self.host:
            raise ValueError("No host was found in event parents. Host must be specified!")


class DictPathEvent(DictEvent):
    def sanitize_data(self, data):
        new_data = dict(data)
        file_blobs = getattr(self.scan, "_file_blobs", False)
        folder_blobs = getattr(self.scan, "_folder_blobs", False)
        blob = None
        try:
            data_path = Path(data["path"])
            if data_path.is_file():
                self.add_tag("file")
                if file_blobs:
                    with open(data_path, "rb") as file:
                        blob = file.read()
            elif data_path.is_dir():
                self.add_tag("folder")
                if folder_blobs:
                    blob = self._tar_directory(data_path)
        except KeyError:
            pass
        if blob:
            new_data["blob"] = base64.b64encode(blob).decode("utf-8")

        return new_data

    def _tar_directory(self, dir_path):
        tar_buffer = io.BytesIO()
        with tarfile.open(fileobj=tar_buffer, mode="w:gz") as tar:
            # Add the entire directory to the tar archive
            tar.add(dir_path, arcname=dir_path.name)
        return tar_buffer.getvalue()


class ASN(DictEvent):
    _always_emit = True
    _quick_emit = True


class CODE_REPOSITORY(DictHostEvent):
    _always_emit = True

    class _data_validator(BaseModel):
        url: str
        _validate_url = field_validator("url")(validators.validate_url)

    def _pretty_string(self):
        return self.data["url"]


class IP_ADDRESS(BaseEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ip = ipaddress.ip_address(self.data)
        self.add_tag(f"ipv{ip.version}")
        if ip.is_private:
            self.add_tag("private-ip")
        self.dns_resolve_distance = getattr(self.parent, "dns_resolve_distance", 0)

    def sanitize_data(self, data):
        return validators.validate_host(data)

    def _host(self):
        return ipaddress.ip_address(self.data)


class DnsEvent(BaseEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # prevent runaway DNS entries
        self.dns_resolve_distance = 0
        parent = getattr(self, "parent", None)
        module = getattr(self, "module", None)
        module_type = getattr(module, "_type", "")
        parent_module = getattr(parent, "module", None)
        parent_module_type = getattr(parent_module, "_type", "")
        if module_type == "DNS":
            self.dns_resolve_distance = getattr(parent, "dns_resolve_distance", 0)
            if parent_module_type == "DNS":
                self.dns_resolve_distance += 1
        # self.add_tag(f"resolve-distance-{self.dns_resolve_distance}")
        # tag subdomain / domain
        if is_subdomain(self.host):
            self.add_tag("subdomain")
        elif is_domain(self.host):
            self.add_tag("domain")
        # tag private IP
        try:
            if self.host.is_private:
                self.add_tag("private-ip")
        except AttributeError:
            pass


class IP_RANGE(DnsEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        net = ipaddress.ip_network(self.data, strict=False)
        self.add_tag(f"ipv{net.version}")

    def sanitize_data(self, data):
        return str(ipaddress.ip_network(str(data), strict=False))

    def _host(self):
        return ipaddress.ip_network(self.data)


class DNS_NAME(DnsEvent):
    def sanitize_data(self, data):
        return validators.validate_host(data)

    def _host(self):
        return self.data

    def _words(self):
        stem = self.host_stem
        if not is_ptr(stem):
            split_stem = stem.split(".")
            if split_stem:
                leftmost_segment = split_stem[0]
                if leftmost_segment == "_wildcard":
                    stem = ".".join(split_stem[1:])
            if stem:
                return extract_words(stem)
        return set()


class OPEN_TCP_PORT(BaseEvent):
    def sanitize_data(self, data):
        return validators.validate_open_port(data)

    def _host(self):
        host, self._port = split_host_port(self.data)
        return host

    def _words(self):
        if not is_ip(self.host) and not is_ptr(self.host):
            return extract_words(self.host_stem)
        return set()


class URL_UNVERIFIED(BaseEvent):
    _status_code_regex = re.compile(r"^status-(\d{1,3})$")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_redirects = getattr(self.parent, "num_redirects", 0)

    def _data_id(self):

        data = super()._data_id()

        # remove the querystring for URL/URL_UNVERIFIED events, because we will conditionally add it back in (based on settings)
        if self.__class__.__name__.startswith("URL") and self.scan is not None:
            prefix = data.split("?")[0]

            # consider spider-danger tag when deduping
            if "spider-danger" in self.tags:
                prefix += "spider-danger"

            if not self.scan.config.get("url_querystring_remove", True) and self.parsed_url.query:
                query_dict = parse_qs(self.parsed_url.query)
                if self.scan.config.get("url_querystring_collapse", True):
                    # Only consider parameter names in dedup (collapse values)
                    cleaned_query = "|".join(sorted(query_dict.keys()))
                else:
                    # Consider parameter names and values in dedup
                    cleaned_query = "&".join(
                        f"{key}={','.join(sorted(values))}" for key, values in sorted(query_dict.items())
                    )
                data = f"{prefix}:{self.parsed_url.scheme}:{self.parsed_url.netloc}:{self.parsed_url.path}:{cleaned_query}"
        return data

    def sanitize_data(self, data):
        self.parsed_url = self.validators.validate_url_parsed(data)

        # special handling of URL extensions
        if self.parsed_url is not None:
            url_path = self.parsed_url.path
            if url_path:
                parsed_path_lower = str(url_path).lower()
                extension = get_file_extension(parsed_path_lower)
                if extension:
                    self.url_extension = extension
                    self.add_tag(f"extension-{extension}")

        # tag as dir or endpoint
        if str(self.parsed_url.path).endswith("/"):
            self.add_tag("dir")
        else:
            self.add_tag("endpoint")

        data = self.parsed_url.geturl()
        return data

    def add_tag(self, tag):
        host_same_as_parent = self.parent and self.host == self.parent.host
        if tag == "spider-danger" and host_same_as_parent and not "spider-danger" in self.tags:
            # increment the web spider distance
            if self.type == "URL_UNVERIFIED":
                self.web_spider_distance += 1
                if self.is_spider_max:
                    self.add_tag("spider-max")
        super().add_tag(tag)

    @property
    def is_spider_max(self):
        if self.scan:
            depth = url_depth(self.parsed_url)
            if (self.web_spider_distance > self.scan.web_spider_distance) or (depth > self.scan.web_spider_depth):
                return True
        return False

    def with_port(self):
        netloc_with_port = make_netloc(self.host, self.port)
        return self.parsed_url._replace(netloc=netloc_with_port)

    def _words(self):
        first_elem = self.parsed_url.path.lstrip("/").split("/")[0]
        if not "." in first_elem:
            return extract_words(first_elem)
        return set()

    def _host(self):
        return make_ip_type(self.parsed_url.hostname)

    @property
    def http_status(self):
        for t in self.tags:
            match = self._status_code_regex.match(t)
            if match:
                return int(match.groups()[0])
        return 0


class URL(URL_UNVERIFIED):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self._dummy and not any(t.startswith("status-") for t in self.tags):
            raise ValidationError(
                'Must specify HTTP status tag for URL event, e.g. "status-200". Use URL_UNVERIFIED if the URL is unvisited.'
            )

    @property
    def resolved_hosts(self):
        # TODO: remove this when we rip out httpx
        return set(".".join(i.split("-")[1:]) for i in self.tags if i.startswith("ip-"))

    @property
    def pretty_string(self):
        return self.data


class STORAGE_BUCKET(DictEvent, URL_UNVERIFIED):
    _always_emit = True
    _suppress_chain_dupes = True

    class _data_validator(BaseModel):
        name: str
        url: str
        _validate_url = field_validator("url")(validators.validate_url)

    def sanitize_data(self, data):
        data = super().sanitize_data(data)
        data["name"] = data["name"].lower()
        return data

    def _words(self):
        return self.data["name"]


class URL_HINT(URL_UNVERIFIED):
    pass


class WEB_PARAMETER(DictHostEvent):

    def _data_id(self):
        # dedupe by url:name:param_type
        url = self.data.get("url", "")
        name = self.data.get("name", "")
        param_type = self.data.get("type", "")
        return f"{url}:{name}:{param_type}"

    def _url(self):
        return self.data["url"]

    def __str__(self):
        max_event_len = 200
        d = str(self.data)
        return f'{self.type}("{d[:max_event_len]}{("..." if len(d) > max_event_len else "")}", module={self.module}, tags={self.tags})'


class EMAIL_ADDRESS(BaseEvent):
    def sanitize_data(self, data):
        return validators.validate_email(data)

    def _host(self):
        data = str(self.data).split("@")[-1]
        host, self._port = split_host_port(data)
        return host

    def _words(self):
        return extract_words(self.host_stem)


class HTTP_RESPONSE(URL_UNVERIFIED, DictEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # count number of consecutive redirects
        self.num_redirects = getattr(self.parent, "num_redirects", 0)
        if str(self.http_status).startswith("3"):
            self.num_redirects += 1

    def _data_id(self):
        return self.data["method"] + "|" + self.data["url"]

    def sanitize_data(self, data):
        url = data.get("url", "")
        self.parsed_url = self.validators.validate_url_parsed(url)
        data["url"] = self.parsed_url.geturl()

        header_dict = {}
        for i in data.get("raw_header", "").splitlines():
            if len(i) > 0 and ":" in i:
                k, v = i.split(":", 1)
                k = k.strip().lower()
                v = v.lstrip()
                if k in header_dict:
                    header_dict[k].append(v)
                else:
                    header_dict[k] = [v]

        data["header-dict"] = header_dict
        # move URL to the front of the dictionary for visibility
        data = dict(data)
        new_data = {"url": data.pop("url")}
        new_data.update(data)

        return new_data

    def _words(self):
        return set()

    def _pretty_string(self):
        return f'{self.data["hash"]["header_mmh3"]}:{self.data["hash"]["body_mmh3"]}'

    @property
    def http_status(self):
        try:
            return int(self.data.get("status_code", 0))
        except (ValueError, TypeError):
            return 0

    @property
    def http_title(self):
        http_title = self.data.get("title", "")
        try:
            return recursive_decode(http_title)
        except Exception:
            return http_title

    @property
    def redirect_location(self):
        location = self.data.get("location", "")
        # if it's a redirect
        if location:
            # get the url scheme
            scheme = is_uri(location, return_scheme=True)
            # if there's no scheme (i.e. it's a relative redirect)
            if not scheme:
                # then join the location with the current url
                location = urljoin(self.parsed_url.geturl(), location)
        return location


class VULNERABILITY(ClosestHostEvent):
    _always_emit = True
    _quick_emit = True
    severity_colors = {
        "CRITICAL": "🟪",
        "HIGH": "🟥",
        "MEDIUM": "🟧",
        "LOW": "🟨",
        "UNKNOWN": "⬜",
    }

    def sanitize_data(self, data):
        self.add_tag(data["severity"].lower())
        return data

    class _data_validator(BaseModel):
        host: Optional[str] = None
        severity: str
        description: str
        url: Optional[str] = None
        path: Optional[str] = None
        _validate_url = field_validator("url")(validators.validate_url)
        _validate_host = field_validator("host")(validators.validate_host)
        _validate_severity = field_validator("severity")(validators.validate_severity)

    def _pretty_string(self):
        return f'[{self.data["severity"]}] {self.data["description"]}'


class FINDING(ClosestHostEvent):
    _always_emit = True
    _quick_emit = True

    class _data_validator(BaseModel):
        host: Optional[str] = None
        description: str
        url: Optional[str] = None
        path: Optional[str] = None
        _validate_url = field_validator("url")(validators.validate_url)
        _validate_host = field_validator("host")(validators.validate_host)

    def _pretty_string(self):
        return self.data["description"]


class TECHNOLOGY(DictHostEvent):
    class _data_validator(BaseModel):
        host: str
        technology: str
        url: Optional[str] = None
        _validate_url = field_validator("url")(validators.validate_url)
        _validate_host = field_validator("host")(validators.validate_host)

    def _data_id(self):
        # dedupe by host+port+tech
        tech = self.data.get("technology", "")
        return f"{self.host}:{self.port}:{tech}"

    def _pretty_string(self):
        return self.data["technology"]


class VHOST(DictHostEvent):
    class _data_validator(BaseModel):
        host: str
        vhost: str
        url: Optional[str] = None
        _validate_url = field_validator("url")(validators.validate_url)
        _validate_host = field_validator("host")(validators.validate_host)

    def _pretty_string(self):
        return self.data["vhost"]


class PROTOCOL(DictHostEvent):
    class _data_validator(BaseModel):
        host: str
        protocol: str
        port: Optional[int] = None
        banner: Optional[str] = None
        _validate_host = field_validator("host")(validators.validate_host)
        _validate_port = field_validator("port")(validators.validate_port)

    def sanitize_data(self, data):
        new_data = dict(data)
        new_data["protocol"] = data.get("protocol", "").upper()
        return new_data

    @property
    def port(self):
        return self.data.get("port", None)

    def _pretty_string(self):
        return self.data["protocol"]


class GEOLOCATION(BaseEvent):
    _always_emit = True
    _quick_emit = True


class PASSWORD(BaseEvent):
    _always_emit = True
    _quick_emit = True


class HASHED_PASSWORD(BaseEvent):
    _always_emit = True
    _quick_emit = True


class USERNAME(BaseEvent):
    _always_emit = True
    _quick_emit = True


class SOCIAL(DictHostEvent):
    _always_emit = True
    _quick_emit = True
    _scope_distance_increment_same_host = True


class WEBSCREENSHOT(DictPathEvent, DictHostEvent):
    _always_emit = True
    _quick_emit = True


class AZURE_TENANT(DictEvent):
    _always_emit = True
    _quick_emit = True


class WAF(DictHostEvent):
    _always_emit = True
    _quick_emit = True

    class _data_validator(BaseModel):
        url: str
        host: str
        waf: str
        info: Optional[str] = None
        _validate_url = field_validator("url")(validators.validate_url)
        _validate_host = field_validator("host")(validators.validate_host)

    def _pretty_string(self):
        return self.data["waf"]


class FILESYSTEM(DictPathEvent):
    pass


class RAW_DNS_RECORD(DictHostEvent, DnsEvent):
    # don't emit raw DNS records for affiliates
    _always_emit_tags = ["target"]


class MOBILE_APP(DictEvent):
    _always_emit = True

    def _pretty_string(self):
        return self.data["url"]


def make_event(
    data,
    event_type=None,
    parent=None,
    context=None,
    module=None,
    scan=None,
    scans=None,
    tags=None,
    confidence=100,
    dummy=False,
    internal=None,
):
    """
    Creates and returns a new event object or modifies an existing one.

    This function serves as a factory for creating new event objects, either by generating a new `Event`
    object or by updating an existing event with additional metadata. If `data` is already an event,
    it updates the event based on the additional parameters provided.

    Parameters:
        data (Union[str, dict, BaseEvent]): The primary data for the event or an existing event object.
        event_type (str, optional): Type of the event, e.g., 'IP_ADDRESS'. Auto-detected if not provided.
        parent (BaseEvent, optional): Parent event leading to this event's discovery.
        context (str, optional): Description of circumstances leading to event's discovery.
        module (str, optional): Module that discovered the event.
        scan (Scan, optional): BBOT Scan object associated with the event.
        scans (List[Scan], optional): Multiple BBOT Scan objects, primarily used for unserialization.
        tags (Union[str, List[str]], optional): Descriptive tags for the event, as a list or a single string.
        confidence (int, optional): Confidence level for the event, on a scale of 1-100. Defaults to 100.
        dummy (bool, optional): Disables data validations if set to True. Defaults to False.
        internal (Any, optional): Makes the event internal if set to True. Defaults to None.

    Returns:
        BaseEvent: A new or updated event object.

    Raises:
        ValidationError: Raised when there's an error in event data or type sanitization.

    Examples:
        If inside a module, e.g. from within its `handle_event()`:
        >>> self.make_event("1.2.3.4", parent=event)
        IP_ADDRESS("1.2.3.4", module=portscan, tags={'ipv4', 'distance-1'})

        If you're outside a module but you have a scan object:
        >>> scan.make_event("1.2.3.4", parent=scan.root_event)
        IP_ADDRESS("1.2.3.4", module=None, tags={'ipv4', 'distance-1'})

        If you're outside a scan and just messing around:
        >>> from bbot.core.event.base import make_event
        >>> make_event("1.2.3.4", dummy=True)
        IP_ADDRESS("1.2.3.4", module=None, tags={'ipv4'})

    Note:
        When working within a module's `handle_event()`, use the instance method
        `self.make_event()` instead of calling this function directly.
    """

    # allow tags to be either a string or an array
    if not tags:
        tags = []
    elif isinstance(tags, str):
        tags = [tags]
    tags = set(tags)

    if is_event(data):
        data = copy(data)
        if scan is not None and not data.scan:
            data.scan = scan
        if scans is not None and not data.scans:
            data.scans = scans
        if module is not None:
            data.module = module
        if parent is not None:
            data.parent = parent
        if context is not None:
            data.discovery_context = context
        if internal == True:
            data.internal = True
        if tags:
            data.tags = tags.union(data.tags)
        event_type = data.type
        return data
    else:
        if event_type is None:
            event_type, data = get_event_type(data)
            if not dummy:
                log.debug(f'Autodetected event type "{event_type}" based on data: "{data}"')

        event_type = str(event_type).strip().upper()

        # Catch these common whoopsies
        if event_type in ("DNS_NAME", "IP_ADDRESS"):
            # DNS_NAME <--> EMAIL_ADDRESS confusion
            if validators.soft_validate(data, "email"):
                event_type = "EMAIL_ADDRESS"
            else:
                # DNS_NAME <--> IP_ADDRESS confusion
                try:
                    data = validators.validate_host(data)
                except Exception as e:
                    log.trace(traceback.format_exc())
                    raise ValidationError(f'Error sanitizing event data "{data}" for type "{event_type}": {e}')
                data_is_ip = is_ip(data)
                if event_type == "DNS_NAME" and data_is_ip:
                    event_type = "IP_ADDRESS"
                elif event_type == "IP_ADDRESS" and not data_is_ip:
                    event_type = "DNS_NAME"
        # USERNAME <--> EMAIL_ADDRESS confusion
        if event_type == "USERNAME" and validators.soft_validate(data, "email"):
            event_type = "EMAIL_ADDRESS"
            tags.add("affiliate")

        event_class = globals().get(event_type, DefaultEvent)

        return event_class(
            data,
            event_type=event_type,
            parent=parent,
            context=context,
            module=module,
            scan=scan,
            scans=scans,
            tags=tags,
            confidence=confidence,
            _dummy=dummy,
            _internal=internal,
        )


def event_from_json(j, siem_friendly=False):
    """
    Creates an event object from a JSON dictionary.

    This function deserializes a JSON dictionary to create a new event object, using the `make_event` function
    for the actual object creation. It sets additional attributes such as the timestamp and scope distance
    based on the input JSON.

    Parameters:
        j (Dict): JSON dictionary containing the event attributes.
                  Must include keys "data" and "type".

    Returns:
        BaseEvent: A new event object initialized with attributes from the JSON dictionary.

    Raises:
        ValidationError: Raised when the JSON dictionary is missing required fields.

    Note:
        The function assumes that the input JSON dictionary is valid and may raise exceptions
        if required keys are missing. Make sure to validate the JSON input beforehand.
    """
    try:
        event_type = j["type"]
        kwargs = {
            "event_type": event_type,
            "scans": j.get("scans", []),
            "tags": j.get("tags", []),
            "confidence": j.get("confidence", 100),
            "context": j.get("discovery_context", None),
            "dummy": True,
        }
        if siem_friendly:
            data = j["data"][event_type]
        else:
            data = j["data"]
        kwargs["data"] = data
        event = make_event(**kwargs)
        event_uuid = j.get("uuid", None)
        if event_uuid is not None:
            event._uuid = uuid.UUID(event_uuid.split(":")[-1])

        resolved_hosts = j.get("resolved_hosts", [])
        event._resolved_hosts = set(resolved_hosts)

        event.timestamp = datetime.datetime.fromisoformat(j["timestamp"])
        event.scope_distance = j["scope_distance"]
        parent_id = j.get("parent", None)
        if parent_id is not None:
            event._parent_id = parent_id
        parent_uuid = j.get("parent_uuid", None)
        if parent_uuid is not None:
            parent_type, parent_uuid = parent_uuid.split(":", 1)
            event._parent_uuid = parent_type + ":" + str(uuid.UUID(parent_uuid))
        return event
    except KeyError as e:
        raise ValidationError(f"Event missing required field: {e}")


def is_event(e):
    return BaseEvent in e.__class__.__mro__



from .base import ModuleTestBase


credshed_auth_response = {
    "access_token": "big_access_token",
    "login": True,
}


credshed_response = {
    "accounts": [
        {
            "e": "bob@blacklanternsecurity.com",
            "h": [],
            "m": "hello my name is bob",
            "p": "",
            "s": [121562],
            "u": "",
        },
        {
            "e": "judy@blacklanternsecurity.com",
            "h": [
                "539FE8942DEADBEEFBC49E6EB2F175AC",
                "D2D8F0E9A4A2DEADBEEF1AC80F36D61F",
                "$2a$12$SHIC49jLIwsobdeadbeefuWb2BKWHUOk2yhpD77A0itiZI1vJqXHm",
            ],
            "m": "hello my name is judy",
            "p": "",
            "s": [80437],
            "u": "",
        },
        {
            "e": "tim@blacklanternsecurity.com",
            "h": [],
            "m": "hello my name is tim",
            "p": "TimTamSlam69",
            "s": [80437],
            "u": "tim",
        },
    ],
    "stats": {
        "accounts_searched": 9820758365,
        "elapsed": "0.00",
        "limit": 1000,
        "query": "blacklanternsecurity.com",
        "query_type": "domain",
        "sources_searched": 129957,
        "total_count": 3,
        "unique_count": 3,
    },
}


class TestCredshed(ModuleTestBase):
    config_overrides = {
        "modules": {"credshed": {"username": "admin", "password": "password", "credshed_url": "https://credshed.com"}}
    }

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url=f"https://credshed.com/api/auth",
            json=credshed_auth_response,
            method="POST",
        )
        module_test.httpx_mock.add_response(
            url=f"https://credshed.com/api/search",
            json=credshed_response,
            method="POST",
        )

    def check(self, module_test, events):
        assert len(events) == 11
        assert 1 == len([e for e in events if e.type == "EMAIL_ADDRESS" and e.data == "bob@blacklanternsecurity.com"])
        assert 1 == len([e for e in events if e.type == "EMAIL_ADDRESS" and e.data == "judy@blacklanternsecurity.com"])
        assert 1 == len([e for e in events if e.type == "EMAIL_ADDRESS" and e.data == "tim@blacklanternsecurity.com"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "HASHED_PASSWORD"
                and e.data == "judy@blacklanternsecurity.com:539FE8942DEADBEEFBC49E6EB2F175AC"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "HASHED_PASSWORD"
                and e.data == "judy@blacklanternsecurity.com:D2D8F0E9A4A2DEADBEEF1AC80F36D61F"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "HASHED_PASSWORD"
                and e.data
                == "judy@blacklanternsecurity.com:$2a$12$SHIC49jLIwsobdeadbeefuWb2BKWHUOk2yhpD77A0itiZI1vJqXHm"
            ]
        )
        assert 1 == len(
            [e for e in events if e.type == "PASSWORD" and e.data == "tim@blacklanternsecurity.com:TimTamSlam69"]
        )
        assert 1 == len([e for e in events if e.type == "USERNAME" and e.data == "tim@blacklanternsecurity.com:tim"])



import re
import yaml

from .base import BaseModule


class secretsdb(BaseModule):
    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["FINDING"]
    flags = ["active", "safe", "web-basic"]
    meta = {
        "description": "Detect common secrets with secrets-patterns-db",
        "created_date": "2023-03-17",
        "author": "@TheTechromancer",
    }
    options = {
        "min_confidence": 99,
        "signatures": "https://raw.githubusercontent.com/blacklanternsecurity/secrets-patterns-db/master/db/rules-stable.yml",
    }
    options_desc = {
        "min_confidence": "Only use signatures with this confidence score or higher",
        "signatures": "File path or URL to YAML signatures",
    }
    deps_pip = ["pyyaml~=6.0"]
    # accept any HTTP_RESPONSE including out-of-scope ones (such as from github_codesearch)
    scope_distance_modifier = 3

    async def setup(self):
        self.rules = []
        self.min_confidence = self.config.get("min_confidence", 99)
        self.sig_file = await self.helpers.wordlist(self.config.get("signatures", ""))
        with open(self.sig_file) as f:
            rules_yaml = yaml.safe_load(f).get("patterns", [])
        for r in rules_yaml:
            r = r.get("pattern", {})
            if not r:
                continue
            name = r.get("name", "").lower()
            confidence = r.get("confidence", "")
            if name and confidence >= self.min_confidence:
                regex = r.get("regex", "")
                try:
                    compiled_regex = re.compile(regex)
                    r["regex"] = compiled_regex
                    self.rules.append(r)
                except Exception:
                    self.debug(f"Error compiling regex: r'{regex}'")
        return True

    async def handle_event(self, event):
        resp_body = event.data.get("body", "")
        resp_headers = event.data.get("raw_header", "")
        all_matches = await self.helpers.run_in_executor(self.search_data, resp_body, resp_headers)
        for matches, name in all_matches:
            matches = [m.string[m.start() : m.end()] for m in matches]
            description = f"Possible secret ({name}): {matches}"
            event_data = {"host": str(event.host), "description": description}
            parsed_url = getattr(event, "parsed_url", None)
            if parsed_url:
                event_data["url"] = parsed_url.geturl()
            await self.emit_event(
                event_data,
                "FINDING",
                parent=event,
                context=f"{{module}} searched HTTP response and found {{event.type}}: {description}",
            )

    def search_data(self, resp_body, resp_headers):
        all_matches = []
        for r in self.rules:
            regex = r["regex"]
            name = r["name"]
            for text in (resp_body, resp_headers):
                if text:
                    matches = list(regex.finditer(text))
                    if matches:
                        all_matches.append((matches, name))
        return all_matches



from .test_module_discord import TestDiscord as DiscordBase


class TestSlack(DiscordBase):
    modules_overrides = ["slack", "excavate", "badsecrets", "httpx"]
    webhook_url = "https://hooks.slack.com/services/deadbeef/deadbeef/deadbeef"
    config_overrides = {"modules": {"slack": {"webhook_url": webhook_url}}}
